/*
* Copyright 1998-2014 VIA Technologies, Inc. All Rights Reserved.
* Copyright 2001-2014 S3 Graphics, Inc. All Rights Reserved.
* Copyright 2013-2014 Shanghai Zhaoxin Semiconductor Co., Ltd. All Rights Reserved.
*
* This file is part of zx.ko
*
* zx.ko is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* zx.ko is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "zx_driver.h"
#include "zx_ioctl.h"
#include "zx_sync.h"
#include "kernel_interface.h"
#include "zx_buffer_object.h"


static int zx_buffer_object_release(struct inode *inode, struct file *filp)
{
    zx_buffer_object_t *bo = filp->private_data;

    filp->private_data = NULL;
    if (bo == NULL)
    {
        return -ENOENT;
    }

    if (bo->read_fence)
    {
        fput(bo->read_fence);
        bo->read_fence = NULL;
    }

    if (bo->write_fence)
    {
        fput(bo->write_fence);
        bo->write_fence = NULL;
    }

    if (bo->lock)
    {
        zx_destroy_mutex(bo->lock);
        bo->lock = NULL;
    }

    zx_free(bo);
    filp->private_data = NULL;

    return 0;
}

static void zx_bo_try_recyle_fence(zx_buffer_object_t *bo)
{
    zx_fence_t *fence;
    struct file *read = NULL;
    struct file *write = NULL;

    zx_mutex_lock(bo->lock);

    if (bo->read_fence != NULL)
    {
        fence = bo->read_fence->private_data;
        if (fence->status == OBJ_SIGNALED)
        {
            read = bo->read_fence;
            bo->read_fence = NULL;
        }
    }
    if (bo->write_fence != NULL)
    {
        fence = bo->write_fence->private_data;
        if (fence->status == OBJ_SIGNALED)
        {
            write = bo->write_fence;
            bo->write_fence = NULL;
        }
    }

    zx_mutex_unlock(bo->lock);

    if (read != NULL)
    {
        fput(read);
    }

    if (write != NULL)
    {
        fput(write);
    }
}

static long zx_bo_set_fence(zx_buffer_object_t *bo, unsigned long arg, struct file **oldp)
{
    int err = 0;
    zx_bo_fence_t    fence = {0};
    struct file *new, *merged, *old;

    if(zx_copy_from_user(&fence, (void __user*)arg, sizeof(zx_bo_fence_t)))
    {
        return -1;
    }

    new = fget(fence.fd);
    if(IS_ERR(new))
    {
        return PTR_ERR(new);
    }

    zx_bo_try_recyle_fence(bo);

    zx_mutex_lock(bo->lock);
    if(*oldp)
    {
        old = *oldp;
        *oldp = NULL;

        merged = zx_sync_fence_merge(old, new);
        if(merged)
        {
            *oldp = merged;
        }
        else
        {
            err = -1;
        }
        zx_mutex_unlock(bo->lock);

        fput(old);
        fput(new);

        zx_mutex_lock(bo->lock);
    }
    else
    {
        *oldp = new;
    }
    zx_mutex_unlock(bo->lock);

    return err;
}

static long zx_bo_get_fence(zx_buffer_object_t *bo, unsigned long arg, struct file **oldp)
{
    int err = 0;
    zx_bo_fence_t    fence = {0};

    zx_bo_try_recyle_fence(bo);

    if(zx_copy_from_user(&fence, (void __user*)arg, sizeof(zx_bo_fence_t)))
    {
        return -1;
    }

    fence.fd = -1;

    zx_mutex_lock(bo->lock);

    if(*oldp)
    {
        err = get_unused_fd_flags(0);

        if(err >= 0)
        {
            fence.fd = err;
            fd_install(fence.fd, *oldp);
            fget(fence.fd);
        }
    }
    zx_mutex_unlock(bo->lock);

    if(err >= 0 && zx_copy_to_user((void __user*)arg, &fence, sizeof(zx_bo_fence_t)))
    {
        err = -1;
    }

    return err >= 0 ? 0 : err;
}

static long zx_buffer_object_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
    long ret = 0;
    zx_buffer_object_t* bo = file->private_data;

    switch(cmd)
    {
    case ZX_IOCTL_GET_READ_FENCE:
        ret = zx_bo_get_fence(bo, arg, &bo->read_fence);
        break;
    case ZX_IOCTL_SET_READ_FENCE:
        ret = zx_bo_set_fence(bo, arg, &bo->read_fence);
        break;
    case ZX_IOCTL_GET_WRITE_FENCE:
        ret = zx_bo_get_fence(bo, arg, &bo->write_fence);
        break;
    case ZX_IOCTL_SET_WRITE_FENCE:
        ret = zx_bo_set_fence(bo, arg, &bo->write_fence);
        break;
    default:
        ret = -EINVAL;
        break;
    }

    return ret;
}

static const struct file_operations zx_buffer_object_fops = {
    .release    = zx_buffer_object_release,
    .unlocked_ioctl = zx_buffer_object_ioctl,
};


int zx_bo_fd_to_handle(zx_file_t *priv, zx_handle_fd_t *hfd)
{
    int ret = 0;
    zx_buffer_object_t *bo;
    struct file *desc = fget(hfd->fd);

    if (IS_ERR(desc))
    {
        zx_error("%s, failed to get file desc of fd(0x%x). \n", __func__, hfd->fd);
        return PTR_ERR(desc);
    }

    if (desc->f_op != &zx_buffer_object_fops)
    {
        ret = -EINVAL;
        goto done;
    }

    bo = desc->private_data;
    hfd->handle = bo->allocation;

done:
    fput(desc);
    return ret;
}

int zx_bo_handle_to_fd(zx_file_t *priv, zx_handle_fd_t *hfd)
{
    int fd = 0;
    zx_card_t *zx = priv->card;
    struct file *bof = krnl_get_bo_file(zx->adapter, hfd->handle);

    if(!bof)
    {
        zx_buffer_object_t *bo = zx_calloc(sizeof(zx_buffer_object_t));
        if (bo == NULL)
        {
            zx_error("%s, fail to allocate buffer object\n", __func__);
            return -ENOMEM;
        }

        bo->adapter = zx->adapter;
        bo->lock = zx_create_mutex();
        bo->allocation = hfd->handle;
        bo->read_fence = NULL;
        bo->write_fence = NULL;

        fd = anon_inode_getfd("zx_buffer_object", &zx_buffer_object_fops, bo, O_RDWR);
        if (fd < 0)
        {
            zx_free(bo);
            return E_FAIL;
        }

        bof = fget(fd);

        krnl_set_bo_file(zx->adapter, hfd->handle, bof);
    }
    else
    {
        fd = get_unused_fd_flags(0);
        if(fd < 0)
        {
            zx_error("%s, fail to get fd for allocation 0x%x\n", __func__, hfd->handle);
            return E_FAIL;
        }

        fd_install(fd, bof);
        fget(fd);
    }

    hfd->fd = fd;

    return 0;
}

