/* 
* Copyright 1998-2014 VIA Technologies, Inc. All Rights Reserved.
* Copyright 2001-2014 S3 Graphics, Inc. All Rights Reserved.
* Copyright 2013-2014 Shanghai Zhaoxin Semiconductor Co., Ltd. All Rights Reserved.
*
* This file is part of zx.ko
* 
* zx.ko is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* zx.ko is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "os_interface.h"
#include "zx_driver.h"
#include "zx_ioctl.h"
#include "zx_cec.h"
#include "zx_debugfs.h"
#include "zx_backlight.h"
#include "zx_version.h"

// assume host's fb is 1920x1080, byte per pixel is 4.
const unsigned int vmem_reserve_for_host = 1920*1080*4;

extern zx_card_t zx_cards[ZX_MAX_CARDS];
extern int zx_run_on_qt;
extern int zx_freezable_patch;

extern void zx_dump_clips(zx_card_t *card);

int zx_add_wait(zx_card_t *zx, zx_file_t *priv, unsigned int wait_events)
{
    unsigned long flags;

    /* do not add empty wait, and one fd only can have one waiter */
    if((wait_events == 0) || (priv->wait_events != 0))
    {
        zx_error("fd have wait event:0x%x, can not add new wait event:0x%x\n", priv->wait_events, wait_events);
        return -1;
    }

    INIT_LIST_HEAD(&priv->wait_list);

    spin_lock_irqsave(&zx->wait_list_lock, flags);

    priv->wait_events = wait_events;

    list_add(&priv->wait_list, &zx->wait_list);

    spin_unlock_irqrestore(&zx->wait_list_lock, flags);

    return 0;
}

int zx_remove_wait(zx_card_t *zx, zx_file_t *priv)
{
    unsigned long flags;

    spin_lock_irqsave(&zx->wait_list_lock, flags);

    list_del(&priv->wait_list);

    priv->wait_events = 0;
    priv->curr_events = 0;

    spin_unlock_irqrestore(&zx->wait_list_lock, flags);

    return 0;
}

void zx_get_event(zx_card_t *zx, zx_file_t *priv, zx_wait_func_t *wait)
{
    unsigned long flags;

    spin_lock_irqsave(&zx->wait_list_lock, flags);

    wait->events     = priv->curr_events;

    priv->curr_events   = 0;

    spin_unlock_irqrestore(&zx->wait_list_lock, flags);
}

int zx_post_event(zx_card_t *zx, unsigned int events)
{
    struct drm_device *dev = pci_get_drvdata(zx->pdev);
    zx_file_t    *priv = NULL;
    unsigned long flags; 

    if (events & ZX_EVENT_VSYNC_0_INTERRUPT)
    {
        ++zx->v_sync_count[0];
        zx_get_nsecs(&zx->v_sync_timestamp[0]);
    }

    if (events & ZX_EVENT_VSYNC_1_INTERRUPT)
    {
        ++zx->v_sync_count[1];
        zx_get_nsecs(&zx->v_sync_timestamp[1]);
    }

    if (events & ZX_EVENT_VSYNC_2_INTERRUPT)
    {
        ++zx->v_sync_count[2];
        zx_get_nsecs(&zx->v_sync_timestamp[2]);
    }

    spin_lock_irqsave(&zx->wait_list_lock, flags);

    list_for_each_entry(priv, &zx->wait_list, wait_list)
    {
        /* only wakeup when wait event happen */
        if(events & priv->wait_events)
        {
            priv->curr_events |= events;
            wake_up_interruptible(&priv->wait_queue);
        }  
    }

    spin_unlock_irqrestore(&zx->wait_list_lock, flags);

    if (events == ZX_EVENT_HOTPLUG)
    {
        drm_sysfs_hotplug_event(dev);
    }

#ifdef NEW_ZXFB
    if(zx->info != NULL)
    {
        if(!zx->enter_vt)
        {
            zxfb_handle_event(zx, events);
        }
        else
        {
            zx->zxfb_pending_events |= events;
        }
    }
#endif  

    return 0;
}

static unsigned int zx_ipc_type_to_event(unsigned int type, int post)
{
    unsigned int event=0;

    if(type == ZX_HDMI_ON_OFF)
    {
        event = post ? ZX_EVENT_HDMI_ON_OFF : ZX_EVENT_HDMI_ON_OFF_FINISH ;
    }
    else if(type == ZX_MONITORES_ON_OFF)
    {
        event = post ? ZX_EVENT_TV_TURN_ON_OFF_MONITORS : ZX_EVENT_TV_TURN_ON_OFF_MONITORS_FINISH ;
    }
    else if(type == ZX_SET_MODE)
    {
        event = post ? ZX_EVENT_TV_SET_MODE : ZX_EVENT_TV_SET_MODE_FINISH;
    }
    
    return event;
}

int zx_ipc_operation(zx_card_t *zx, zx_file_t *priv, zx_ipc_operation_t *ipc_operation)
{
    int                     ret      = 0;
    unsigned long           timeout  = 0;
    long                    wait_ret = 0;

    unsigned int            reply_event = 0, post_event = 0;

    zx_assert(ipc_operation->type < ZX_IPC_TYPE_COUNT);

    zx_assert(ipc_operation->size_in < MAX_IPC_OPERATION_DATA*4);
    zx_assert(ipc_operation->size_out < MAX_IPC_OPERATION_DATA*4);

    switch(ipc_operation->operation)
    {
        case ZX_IPC_GET:
            if(ipc_operation->send_data != NULL)
            {
                ret = zx_copy_to_user(ipc_operation->send_data, zx->ipc_data[ipc_operation->type].send_data, zx->ipc_data[ipc_operation->type].size_in);
            }

            break;

        case ZX_IPC_SEND:
            if(ipc_operation->send_data != NULL)
            {
                ret = zx_copy_from_user(zx->ipc_data[ipc_operation->type].send_data, ipc_operation->send_data, ipc_operation->size_in);
                zx->ipc_data[ipc_operation->type].size_in = ipc_operation->size_in;
            }

            reply_event = zx_ipc_type_to_event(ipc_operation->type, 0);
            post_event  = zx_ipc_type_to_event(ipc_operation->type, 1);

            if(zx_add_wait(zx, priv, reply_event))
            {
                ret = -1;
                break;
            }

            zx_post_event(zx, post_event);

            timeout = msecs_to_jiffies(ipc_operation->timeout);

            zx_assert(timeout > 0);

            wait_ret = wait_event_interruptible_timeout(priv->wait_queue, reply_event & priv->curr_events, timeout);

            zx_remove_wait(zx, priv);

            if(wait_ret > 0 && ipc_operation->reply_data != NULL)
            {
                if(zx->ipc_data[ipc_operation->type].size_out != ipc_operation->size_out)
                {
                    zx_error("IPC type:%d need reply size:%d, but actual size:%d, return:0x%x\n", 
                            ipc_operation->type, ipc_operation->size_out, zx->ipc_data[ipc_operation->type].size_out, wait_ret);
                }

                zx_assert(zx->ipc_data[ipc_operation->type].size_out == ipc_operation->size_out);
                ret = zx_copy_to_user(ipc_operation->reply_data, zx->ipc_data[ipc_operation->type].reply_data, zx->ipc_data[ipc_operation->type].size_out);
            }else
            {
                zx_info("ipc wait 0x%x timeout, wait return 0x%x\n", reply_event, wait_ret);
                ret = -1;
            }
            break;

        case ZX_IPC_REPLY:
            if(ipc_operation->reply_data != NULL)
            {
                ret = zx_copy_from_user(zx->ipc_data[ipc_operation->type].reply_data, ipc_operation->reply_data, ipc_operation->size_out);
                zx->ipc_data[ipc_operation->type].size_out = ipc_operation->size_out;
            }

            reply_event = zx_ipc_type_to_event(ipc_operation->type, 0);

            zx_post_event(zx, reply_event);

            break;
        default:
            zx_info("no this ipc operation\n"); 
            zx_assert(0);
    }

    return ret;
}

pgprot_t os_get_pgprot_val(unsigned int *cache_type, pgprot_t old_prot, int io_map)
{
    pgprot_t prot = old_prot;

    if(*cache_type == ZX_MEM_UNCACHED)
    {   
        prot = pgprot_noncached(old_prot);
    }
    else if(*cache_type == ZX_MEM_WRITE_COMBINED)
    {
#ifdef CONFIG_X86
#ifdef CONFIG_X86_PAT
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,19,0)
        prot = __pgprot((pgprot_val(old_prot) & ~ _PAGE_CACHE_MASK) | cachemode2protval(_PAGE_CACHE_MODE_WC));
#else
        prot = __pgprot((pgprot_val(old_prot) & ~ _PAGE_CACHE_MASK) | _PAGE_CACHE_WC);
#endif
#else
        prot = pgprot_noncached(old_prot);

        *cache_type = ZX_MEM_UNCACHED;	
#endif
#else
        prot = pgprot_writecombine(old_prot);
#endif
        
    }
#ifdef ZX_AHB_BUS
    else if(*cache_type == ZX_MEM_WRITE_BACK)
    {   
        // for snoop path, current linux kennel defalut is not WA+WB
        prot = __pgprot((pgprot_val(old_prot)&(~(L_PTE_MT_MASK))) | (L_PTE_MT_WRITEBACK | L_PTE_MT_WRITEALLOC));
    }
#endif
    return prot;
}

#ifndef __frv__
#if 0
int zx_map_system_ram(struct vm_area_struct* vma, zx_map_argu_t *map)
{
    unsigned long start = vma->vm_start;

    unsigned long i = 0, pfn;

    unsigned int  cache_type;

    struct os_pages_memory *memory = map->memory;
    
#if LINUX_VERSION_CODE <= 0x02040e
    vma->vm_flags |= VM_LOCKED;
#elif LINUX_VERSION_CODE < 0x030700
    vma->vm_flags |= VM_RESERVED;
#else
    vma->vm_flags |= VM_DONTEXPAND | VM_DONTDUMP;
#endif

    cache_type = map->flags.cache_type;

#if LINUX_VERSION_CODE >= 0x020612
    vma->vm_page_prot = vm_get_page_prot(vma->vm_flags);
#endif
    vma->vm_page_prot = os_get_pgprot_val(&cache_type, vma->vm_page_prot, 0);

    for(i = 0; i < memory->page_count; i++)
    {
        pfn = page_to_pfn(memory->pages[i]);

        if(remap_pfn_range(vma, start, pfn, PAGE_SIZE, vma->vm_page_prot))
        {
            return -EAGAIN;
        }

        start += PAGE_SIZE;
    }

    return 0;
}
#endif
int zx_map_system_io(struct vm_area_struct* vma, zx_map_argu_t *map)
{
    unsigned int  cache_type = map->flags.cache_type;
#if LINUX_VERSION_CODE >= 0x020612
    vma->vm_page_prot = vm_get_page_prot(vma->vm_flags);
#endif
    vma->vm_page_prot = os_get_pgprot_val(&cache_type, vma->vm_page_prot, 1);

    map->flags.cache_type = cache_type;

    if (remap_pfn_range(vma, vma->vm_start,
                vma->vm_pgoff,
                vma->vm_end - vma->vm_start,
                vma->vm_page_prot))
    {
        return -EAGAIN;
    }

    return 0;    
}
#if 1
int zx_map_system_ram(struct vm_area_struct* vma, zx_map_argu_t *map)
{
    struct os_pages_memory *memory = map->memory;
    struct os_pmem_area    *area   = NULL;
    struct os_pmem_block   *block  = NULL;
	
    unsigned long start = vma->vm_start;
    unsigned long pfn;
    unsigned int  cache_type, curr_block_offset;
	
    int mapped_size = PAGE_ALIGN(map->size);
    int offset      = _ALIGN_DOWN(map->offset, PAGE_SIZE);
    int area_offset = offset % memory->area_size;
    int range_size, left_size  = 0;
    
#if LINUX_VERSION_CODE <= 0x02040e
    vma->vm_flags |= VM_LOCKED;
#elif LINUX_VERSION_CODE < 0x030700
    vma->vm_flags |= VM_RESERVED;
#else
    vma->vm_flags |= VM_DONTEXPAND | VM_DONTDUMP;
#endif

    cache_type = map->flags.cache_type;

#if LINUX_VERSION_CODE >= 0x020612
    vma->vm_page_prot = vm_get_page_prot(vma->vm_flags);
#endif
    vma->vm_page_prot = os_get_pgprot_val(&cache_type, vma->vm_page_prot, 0);
	
    area  = &memory->areas[offset / memory->area_size];
    block = area->block;

    curr_block_offset = area_offset + area->start;

    left_size = mapped_size;

    while(left_size > 0)
    {
        if(curr_block_offset >= block->size)
        {
            curr_block_offset -= block->size;
            block++;

            if(block - memory->blocks >= memory->block_cnt)
            {
                zx_error("SOMETHING WRONG. block overflow.\n");

                zx_assert(0);
            }
        }

        if(curr_block_offset >= block->size)
        {
            zx_error("SOMETHING WRONG. block offset excceed block size.\n");

            zx_assert(0);
        }

        range_size = (block->size - curr_block_offset) > left_size ? 
                     left_size : 
                     block->size - curr_block_offset;

        pfn  = page_to_pfn(block->pages);
        pfn += curr_block_offset / PAGE_SIZE;

        if(remap_pfn_range(vma, start, pfn, range_size, vma->vm_page_prot))
        {
            return -EAGAIN;
        }

        start             += range_size;
        left_size         -= range_size;
        curr_block_offset += range_size;   
    }

    return 0;
}
#endif
#endif

static irqreturn_t zx_isr(int irq, void *data)
{
    zx_card_t *zx = data;

    if((zx->isr_func)(zx->adapter))
    {
        tasklet_schedule(&zx->tasklet);
    }
    
    return IRQ_HANDLED;
}

void zx_interrupt_init(zx_card_t *zx)
{
    /* request isr */
    krnl_get_isr_func(zx->adapter, (void**)&zx->isr_func, (void**)&zx->isr_bf_func);

    krnl_set_callback_func(zx->adapter, OS_CALLBACK_POST_EVENT, zx_post_event, zx);

    tasklet_init(&zx->tasklet, (void*)zx->isr_bf_func, (unsigned long)zx->adapter);
    tasklet_disable(&zx->tasklet);
    tasklet_enable(&zx->tasklet); 

    if(zx_register_interrupt(zx, zx_isr))
    {
        zx_error("zx_regster_interrupt failed.\n");
    }
    krnl_enable_interrupt(zx->adapter);
}

void zx_interrupt_reinit(zx_card_t *zx)
{
    /* request isr */
    krnl_get_isr_func(zx->adapter, (void**)&zx->isr_func, (void**)&zx->isr_bf_func);

    krnl_set_callback_func(zx->adapter, OS_CALLBACK_POST_EVENT, zx_post_event, zx);

    tasklet_enable(&zx->tasklet); 

    if(zx_register_interrupt(zx, zx_isr))
    {
        zx_error("zx_regster_interrupt failed.\n");
    }
}

void zx_interrupt_deinit(zx_card_t *zx)
{
    krnl_disable_interrupt(zx->adapter);

    zx_unregister_interrupt(zx);

    tasklet_disable(&zx->tasklet);
}

static krnl_import_func_list_t zx_export =
{
    .udelay                        = zx_udelay,
    .begin_timer                   = zx_begin_timer,
    .end_timer                     = zx_end_timer,
    .do_div                        = zx_do_div,
    .msleep                        = zx_msleep,
    .assert                        = zx_assert,
    .copy_from_user                = zx_copy_from_user,
    .copy_to_user                  = zx_copy_to_user,
    .slowbcopy_tobus               = zx_slowbcopy_tobus,
    .slowbcopy_frombus             = zx_slowbcopy_frombus,
    .memset                        = zx_memset,
    .memcpy                        = zx_memcpy,
    .memcmp_priv                   = zx_memcmp,
    .byte_copy                     = zx_byte_copy,
    .strcmp                        = zx_strcmp,
    .strcpy                        = zx_strcpy,
    .strncmp                       = zx_strncmp,
    .strncpy                       = zx_strncpy,
    .strlen                        = zx_strlen,
    .secure_read32                 = zx_secure_read32,
    .read32                        = zx_read32,
    .read16                        = zx_read16,
    .read8                         = zx_read8,
    .secure_write32                = zx_secure_write32,
    .write32                       = zx_write32,
    .write16                       = zx_write16,
    .write8                        = zx_write8,
    .inb                           = zx_inb,
    .outb                          = zx_outb,
    .memsetio                      = zx_memsetio,
    .memcpy_fromio                 = zx_memcpy_fromio,
    .memcpy_toio                   = zx_memcpy_toio,
    .file_open                     = zx_file_open,
    .file_close                    = zx_file_close,
    .file_read                     = zx_file_read,
    .file_write                    = zx_file_write,
    .vsprintf                      = zx_vsprintf,
    .vsnprintf                     = zx_vsnprintf,
    .sscanf                        = zx_sscanf,
    .printk                        = zx_printk,
    .cb_printk                     = zx_cb_printk,
    .seq_printf                    = zx_seq_printf,
    .find_first_zero_bit           = zx_find_first_zero_bit,
    .find_next_zero_bit            = zx_find_next_zero_bit,
    .set_bit                       = zx_set_bit,
    .clear_bit                     = zx_clear_bit,
    .getsecs                       = zx_getsecs,
    .get_nsecs                     = zx_get_nsecs,
    .create_thread                 = zx_create_thread,
    .destroy_thread                = zx_destroy_thread,
    .thread_should_stop            = zx_thread_should_stop,
    .create_atomic                 = zx_create_atomic,
    .destroy_atomic                = zx_destroy_atomic,
    .atomic_read                   = zx_atomic_read,
    .atomic_inc                    = zx_atomic_inc,
    .atomic_dec                    = zx_atomic_dec,
    .create_mutex                  = zx_create_mutex,
    .destroy_mutex                 = zx_destroy_mutex,
    .mutex_lock                    = zx_mutex_lock,
    .mutex_lock_killable           = zx_mutex_lock_killable,
    .mutex_trylock                 = zx_mutex_trylock,
    .mutex_unlock                  = zx_mutex_unlock,
    .create_sema                   = zx_create_sema,
    .destroy_sema                  = zx_destroy_sema,
    .down                          = zx_down,
    .down_trylock                  = zx_down_trylock,
    .up                            = zx_up,
    .create_rwsema                 = zx_create_rwsema,
    .destroy_rwsema                = zx_destroy_rwsema,
    .down_read                     = zx_down_read,
    .down_write                    = zx_down_write,
    .up_read                       = zx_up_read,
    .up_write                      = zx_up_write,
    .create_spinlock               = zx_create_spinlock,
    .destroy_spinlock              = zx_destroy_spinlock,
    .spin_lock                     = zx_spin_lock,
    .spin_unlock                   = zx_spin_unlock,
    .spin_lock_irqsave             = zx_spin_lock_irqsave,
    .spin_unlock_irqrestore        = zx_spin_unlock_irqrestore,
    .create_event                  = zx_create_event,
    .destroy_event                 = zx_destroy_event,
    .wait_event_thread_safe        = zx_wait_event_thread_safe,
    .wait_event                    = zx_wait_event,
    .wake_up_event                 = zx_wake_up_event,
    .thread_wait                   = zx_thread_wait,
    .create_wait_queue         = zx_create_wait_queue,
    .wait_for_events             = thread_wait_for_events,
    .thread_wake_up                = zx_thread_wake_up,
    ._thread_wake_up              = general_thread_wake_up,
    .dump_stack                    = zx_dump_stack,
    .try_to_freeze                 = zx_try_to_freeze,
    .freezable                     = zx_freezable,
    .clear_freezable               = zx_clear_freezable,
    .set_freezable                 = zx_set_freezable,
    .freezing                      = zx_freezing,
    .get_current_pid               = zx_get_current_pid,
    .get_current_tid               = zx_get_current_tid,
    .flush_cache                   = zx_flush_cache,
    .inv_cache                     = zx_inv_cache,
    .get_page_phys_address         = zx_get_page_phys_address,
    .ioremap                       = zx_ioremap,
    .iounmap_priv                  = zx_iounmap,
    .mtrr_add                      = zx_mtrr_add,
    .mtrr_del                      = zx_mtrr_del,
    .get_mem_info                  = zx_get_mem_info,
    .register_shrinker             = zx_register_shrinker,
    .unregister_shrinker           = zx_unregister_shrinker,
    .pages_memory_swapout          = zx_pages_memory_swapout,
    .pages_memory_swapin           = zx_pages_memory_swapin,
    .release_file_storage          = zx_release_file_storage,
    .get_bus_config                = zx_get_bus_config,
    .get_platform_config           = zx_get_platform_config,
    .get_command_status16          = zx_get_command_status16,
    .write_command_status16        = zx_write_command_status16,
    .get_command_status32          = zx_get_command_status32,
    .write_command_status32        = zx_write_command_status32,
    .get_rom_start_addr            = zx_get_rom_start_addr,
    .get_rom_save_addr             = zx_get_rom_save_addr,
    .write_rom_save_addr           = zx_write_rom_save_addr,
    .get_bar1                      = zx_get_bar1,
    //gpio
    .gpio_set                      = zx_gpio_setvalue,
    .gpio_get                      = zx_gpio_getvalue,
    .gpio_request                  = zx_gpio_request,
    .gpio_free                     = zx_gpio_free,
    .gpio_direction_input          = zx_gpio_direction_input,
    .gpio_direction_output         = zx_gpio_direction_output,
    //regulator
    .regulator_get                 = zx_regulator_get,
    .regulator_enable              = zx_regulator_enable,
    .regulator_disable             = zx_regulator_disable,
    .regulator_is_enabled          = zx_regulator_is_enabled,
    .regulator_get_voltage         = zx_regulator_get_voltage,
    .regulator_set_voltage         = zx_regulator_set_voltage,
    .regulator_put                 = zx_regulator_put,

    .malloc_track                  = zx_malloc_track,
    .calloc_track                  = zx_calloc_track,
    .free_track                    = zx_free_track,
    .malloc_priv                   = zx_malloc_priv,
    .calloc_priv                   = zx_calloc_priv,
    .free_priv                     = zx_free_priv,
    .allocate_pages_memory_track   = zx_allocate_pages_memory_track,
    .free_pages_memory_track       = zx_free_pages_memory_track,
    .allocate_pages_memory_priv    = zx_allocate_pages_memory_priv,
    .free_pages_memory_priv        = zx_free_pages_memory_priv,
    .map_pages_memory_track        = zx_map_pages_memory_track,
    .unmap_pages_memory_track      = zx_unmap_pages_memory_track,
    .map_pages_memory_priv         = zx_map_pages_memory_priv,
    .unmap_pages_memory_priv       = zx_unmap_pages_memory_priv,
    .map_io_memory_track           = zx_map_io_memory_track,
    .unmap_io_memory_track         = zx_unmap_io_memory_track,
    .map_io_memory_priv            = zx_map_io_memory_priv,
    .unmap_io_memory_priv          = zx_unmap_io_memory_priv,
    .acquire_pfns                  = zx_acquire_pfns,
    .release_pfns                  = zx_release_pfns,
#ifdef ZX_TRACE_EVENT
    .register_trace_events         = zx_register_trace_events,
    .unregister_trace_events       = zx_unregister_trace_events,
    .fence_back_trace_event        = zx_fence_back_trace_event,
    .cmd_render_trace_event        = zx_cmd_render_trace_event,
    .dma_submit_trace_event        = zx_dma_submit_trace_event,
#endif

#ifdef ENABLE_TS
    .register_gfx_ts_handle        = zx_register_gfx_ts_handle,
#else
    .register_gfx_ts_handle        = NULL,
#endif

    .query_platform_caps           = zx_query_platform_caps,

    .dma_map_page                  = zx_dma_map_page,
    .dma_unmap_page                = zx_dma_unmap_page,
    .get_page                      = zx_get_page,
    .get_page_size                 = zx_get_page_size,
    .get_page_dma_addr             = zx_get_page_dma_addr,
    .set_page_dma_addr             = zx_set_page_dma_addr,
    .dma_mapping_error             = zx_dma_mapping_error,
    .fetch_iov_addr                = zx_fetch_iov_addr,
    .console_lock                  = zx_console_lock,
    .console_unlock                = zx_console_unlock,

    .sg_alloc_table                = zx_sg_alloc_table,
    .sg_set_pages                  = zx_sg_set_pages,
    .dma_map_sg                    = zx_dma_map_sg,
    .pages_memory_for_each_continues = zx_pages_memory_for_each_continues,
};

/***/
#include <linux/proc_fs.h>

#ifdef NO_PROC_CREATE_FUNC
static int zx_read_proc(char *page, char **start, off_t off,
			   int count, int *eof, void *data)
{

    zx_card_t    *zx = data;

    krnl_dump_resource(zx->adapter, 5, 0);

    return 0;
}
#else
static ssize_t zx_proc_read(struct file *filp, char *buf, size_t count, loff_t *offp)
{
    zx_card_t *zx = PDE_DATA(file_inode(filp));

    krnl_dump_resource(zx->adapter, 1, 0);

    return 0;
}

#define ZX_DUMP_CLIPS_INDEX           0x300

#define ZX_DUMP_POWER_STATUS_INDEX    0x1000
#define ZX_DUMP_SENSOR_TEMP_INDEX     0x2000
#define ZX_FORCE_DVFS_INDEX           0x3000
#define ZX_FORCE_DVFS_VIDEO_INDEX     0x4000

#define ZX_DUMP_COMPRESS_PS_INDEX    (0x100 + ZX_STREAM_PS)
#define ZX_DUMP_COMPRESS_SS_INDEX    (0x100 + ZX_STREAM_SS)
#define ZX_DUMP_COMPRESS_TS_INDEX    (0x100 + ZX_STREAM_TS)
#define ZX_DUMP_COMPRESS_4S_INDEX    (0x100 + ZX_STREAM_4S)
#define ZX_DUMP_COMPRESS_5S_INDEX    (0x100 + ZX_STREAM_5S)
#define ZX_DUMP_COMPRESS_6S_INDEX    (0x100 + ZX_STREAM_6S)
#define ZX_DUMP_COMPRESS_MAX_INDEX   (0x100 + ZX_STREAM_MAX)


static ssize_t zx_proc_write(struct file *filp, const char __user *buf, size_t count, loff_t *offp)
{
    zx_card_t *zx = PDE_DATA(file_inode(filp));

    char mode = '0';
    int dump_index = 0;
    unsigned int ret = 0;
    char temp[20] = {0};
    unsigned int interval = 0, crtc = 0;
    unsigned long value = 0;

    if(count > 1 && count < 20)
    {
        ret = copy_from_user(&temp[0], buf, count);
        temp[count-1] = '\0';

        if(count == 2)
        {
            mode = temp[0];

            if(mode >= '0' && mode <= '9')
            {
                dump_index = (mode - '0');
            }
            else if((mode >= 'a' && mode < 'e'))
            {
                dump_index =  mode - 'a' + 0xa;
            }
            else if(mode == 'p')
            {
                dump_index = ZX_DUMP_POWER_STATUS_INDEX;
            }
            else if(mode == 'k')
            {
                dump_index = ZX_DUMP_CLIPS_INDEX;
            }
            else
            {
               zx_error("echo error proc mode, should between 0 and d \n");
               return -EFAULT;
            }
        }
        else
        {
            value = simple_strtoul(&temp[1], NULL, 16);

            zx_info("value %d \n", value);

            //input format should be: [T|t][0~9]
            if(temp[0] == 't' || temp[0] == 'T')
            {
                dump_index = ZX_DUMP_SENSOR_TEMP_INDEX;
                interval = value;
            }
            else if(temp[0] == 'f' || temp[0] == 'F')//set dvfs clamp
            {
                dump_index = ZX_FORCE_DVFS_INDEX;
            }
            else if(temp[0] == 'v' || temp[0] == 'V')//force video index
            {
                dump_index = ZX_FORCE_DVFS_VIDEO_INDEX;
            }
            else if(temp[0] == 'i' || temp[0] == 'I')
            {
                zx_begin_end_isp_dvfs(0, value);
            }
            else if((temp[0] == 'c' || temp[0] == 'C') &&
                    (temp[2] == 's' || temp[2] == 'S'))
            {
                if(zx_strlen(temp) != 4)
                {
                    zx_info("invalid flag buffer dump parameter!\n");
                }

                crtc = simple_strtoul(&temp[1], NULL, 16);
                dump_index = simple_strtoul(&temp[3], NULL, 16);

                dump_index += ZX_DUMP_COMPRESS_PS_INDEX;
            }
            else
            {
               zx_error("invalid index\n");
            }
         }
    }
    else
    {
        zx_error("proc write pass too long param, should less then 20\n");
    }

    if (dump_index == ZX_DUMP_CLIPS_INDEX)
    {
        zx_dump_clips(zx);
    }
    else if((dump_index >=0 && dump_index <=5) || dump_index == ZX_DUMP_POWER_STATUS_INDEX)
    {
        krnl_dump_resource(zx->adapter, dump_index, 0);
    }
    else if(dump_index > 5 && dump_index < 0xe )
    {
        krnl_save_force_dvfs_index(zx->adapter, dump_index);
    }
    else if(dump_index == ZX_DUMP_SENSOR_TEMP_INDEX)
    {
        krnl_dump_sensor_temprature(zx->adapter, interval);
    }
    else if(dump_index == ZX_FORCE_DVFS_INDEX)
    {
        krnl_force_dvfs_index(zx->adapter, value);
    }
    else if(dump_index == ZX_FORCE_DVFS_VIDEO_INDEX)
    {
        krnl_force_dvfs_video_index(zx->adapter, value);
    }
    else if((dump_index >= ZX_DUMP_COMPRESS_PS_INDEX)&&
            (dump_index <= ZX_DUMP_COMPRESS_MAX_INDEX))
    {
        krnl_dump_resource(zx->adapter, dump_index, crtc);
    }

    return count;
}

typedef enum _ZX_CBIOS_DUMP_TYPE   {
    ZX_CBIOS_DUMP_NONE            = 0x00,
    ZX_CBIOS_DUMP_EDID_INFO       = 0x01,
    ZX_CBIOS_DUMP_MODELIST_INFO   = 0x02,
    ZX_CBIOS_DUMP_REGISTER_INFO   = 0x04,
    ZX_CBIOS_DUMP_MODE_INFO       = 0x08,
    ZX_CBIOS_DUMP_VCP_INFO        = 0x10,
    ZX_CBIOS_DUMP_CLOCK_INFO      = 0x20,
    ZX_CBIOS_DUMP_SHADOW_INFO     = 0x40,
    ZX_CBIOS_DUMP_TYPE_MASK       = 0XFF,
}ZX_CBIOS_DUMP_TYPE;

typedef enum _ZX_CBIOS_PROC_TYPE   {
	ZX_CBIOS_NONE          =0x00,
    	ZX_CBIOS_DUMP           =0x01,
	ZX_CBIOS_RW               =0x02,
	ZX_CBIOS_TYPE_MASK = 0xff,
}ZX_CBIOS_PROC_TYPE;




static void print_help(void)
{
    zx_info("[=======================]\n\n");
    zx_info("0: dump EDID Info;\n");
    zx_info("1: dump Mode List;\n");
    zx_info("2: dump Register;\n");
    zx_info("3: dump Mode Info;\n");
    zx_info("4: dump VCP Info;\n");
    zx_info("5: dump CLK Info;\n");
    zx_info("6: dump Shadow Info;\n");
    zx_info("write byte/dword: eb/ed [mmio offset] [value];\n");
    zx_info("read byte/dword: db/dd [mmio offset];\n");	
    zx_info("[=======================]\n\n");
    return;
}
static ssize_t cbios_proc_read(struct file *filp, char *buf, size_t count, loff_t *offp)
{
    zx_card_t *zx = PDE_DATA(file_inode(filp));
    zx_query_info_t queryinfo = {0};
    int ret = -1;
    cbiosProcReadInfo ReadInfo = {0};
	
    queryinfo.type =  ZX_QUERY_VBIOS_VERSION;    
    krnl_query_info(zx->adapter,&queryinfo);
    ReadInfo.vbiosVersion = queryinfo.value;

    zx_memset(&queryinfo,0,sizeof(zx_query_info_t));
    queryinfo.type = ZX_QUERY_ENGINE_CLOCK;    
    krnl_query_info(zx->adapter,&queryinfo);
    ReadInfo.Eclk = (queryinfo.value + 5000)/10000;

    zx_memset(&queryinfo,0,sizeof(zx_query_info_t));
    queryinfo.type = ZX_QUERY_I_CLOCK;    
    krnl_query_info(zx->adapter,&queryinfo);
    ReadInfo.Iclk = (queryinfo.value + 5000)/10000;

    zx_memset(&queryinfo,0,sizeof(zx_query_info_t));
    queryinfo.type = ZX_QUERY_VIDEO_CLOCK;    
    krnl_query_info(zx->adapter,&queryinfo);
    ReadInfo.Vclk = (queryinfo.value + 5000)/10000;

    zx_memset(&queryinfo,0,sizeof(zx_query_info_t));
    queryinfo.type = ZX_QUERY_DEVICE_ID;    
    krnl_query_info(zx->adapter,&queryinfo);
    ReadInfo.DeviceId  = queryinfo.value;
    if( ReadInfo.DeviceId  == 0x3A04)
    {     
    	zx_memset(&queryinfo,0,sizeof(zx_query_info_t));
        queryinfo.type = ZX_QUERY_VCP_CLOCK;    
        krnl_query_info(zx->adapter,&queryinfo);
        ReadInfo.VcpClk = (queryinfo.value + 5000)/10000;
        ReadInfo.VcpClk /= 2;
    }
	ReadInfo.DrvVersion = DRIVER_VERSION;
	zx_strcpy(ReadInfo.ReleaseDate,DRIVER_DATE);
	

    if(!copy_to_user(buf,&ReadInfo,sizeof(cbiosProcReadInfo)))
    {   
        return sizeof(cbiosProcReadInfo);
    }
	return ret;
}

static ssize_t cbios_proc_write(struct file *filp, const char __user *buf, size_t count, loff_t *offp)
{
    zx_card_t *     zx = PDE_DATA(file_inode(filp));
    char             mode = '0';
    int              dump_index = 0;
    unsigned int     ret = 0;
    char*            temp = NULL;
    char*            pointer = NULL;	
    unsigned int     type = ZX_CBIOS_NONE;
    unsigned long    address = 0;
    char*            end_address;	
    unsigned int     value = 0;
    unsigned int     bwrite = 0;
    unsigned int     size = 0;	
    unsigned int     dump_type = ZX_CBIOS_DUMP_NONE;
    unsigned int     i = 0;	

    temp = zx_calloc(count);
    ret = copy_from_user(temp, buf, count);
    temp[count-1] = '\0';
    pointer = temp;	

    if(count == 2)
    {
        type = ZX_CBIOS_DUMP;

        mode = pointer[0];
            
        if(mode == 'h' || mode == 'H')
        {
            type = ZX_CBIOS_NONE;
            print_help();
        }
        else if(mode >= '0' && mode <= '6')
        {
            dump_index = (mode - '0');
        }
        else
        {
           zx_error("echo error proc mode, should between 0 and 6 \n");
           goto fail;
        }
    }
    else if(count > 2)
    {
        type = ZX_CBIOS_RW;
        if((pointer[0] == 'e' && pointer[1] == 'b') || (pointer[0] == 'e' && pointer[1] == 'd'))
        {
            bwrite = 1;
	        i = 0;
	        if(pointer[1] == 'b')
	        {
	            size = sizeof(unsigned char);
	        }
	        else
	        {
		        size= sizeof(unsigned int);	
	        }
	        pointer= (pointer+2);
            while(pointer[i] == ' ')    //skip space
            {
	            i++;
            }
	        address = simple_strtoul(&pointer[i], &end_address, 0);
	        while(*end_address == ' ')   //skip space
	        {
	            end_address ++;
	        }
	        value = (unsigned int)simple_strtoul(end_address, NULL, 0); 
        }
        else if((pointer[0] == 'd' && pointer[1] == 'b') || (pointer[0] == 'd' && pointer[1] == 'd'))
        {
            bwrite = 0;
	        i = 0;	
	        if(pointer[1] == 'b')
	        {
	            size = sizeof(unsigned char);
	        }
	        else
	        {
		        size= sizeof(unsigned int);	
	        }
	        pointer= pointer+2;         
            while(pointer[i] == ' ')    //skip space
            {
	            i++;
            }
	        address = simple_strtoul(&pointer[i], NULL, 0);	 
        }
	    else
	    {
	        zx_error("invalid format\n");
	        goto fail; 	 
	    }
    }
    else
    {
        zx_error("invalid format\n");
        goto fail;
    }

    if(type == ZX_CBIOS_DUMP)
    {
        dump_type = 1 << dump_index;
        krnl_dump_cbios_info(zx->adapter, dump_type);
    }
    else if(type == ZX_CBIOS_RW)
    {
        if(bwrite)
        {
            krnl_mmio_write(zx->adapter, (void*)address, size, value);
        }
        else
        {
	        value = krnl_mmio_read(zx->adapter, (void*)address, size);
        }
	    zx_info("%s %x%s %x\n", bwrite? "Write": "Read", address, bwrite? "" : " =", value);	
    }
    else
    {
        zx_info("nothing to be done!\n");
    }

    zx_free(temp);
    return count;

fail:
    zx_free(temp);
    return -EFAULT;	
}

static void zx_aux_help(void)
{
    zx_info("aux use help\n");
    zx_info("native read: native -o deviceid -r address  -l length\n");
    zx_info("native write:native -o deviceid -w address -l length -d data(0) data(1) ... data(length-1)\n");
    zx_info("i2c over aux read: i2c -o deviceid -r address -l length\n");
    return;
}

static ssize_t zx_aux_proc_write(struct file *filp, const char __user *buf, size_t count, loff_t *offp)
{
    zx_card_t *   zx = PDE_DATA(file_inode(filp));
    zx_aux_t       aux = {0};

    char*            temp = NULL;
    unsigned int i, j,NumOfStrings;
    char             **str = NULL;
    char             *token = NULL;
    char             filter[] = " ,\t\n";
    unsigned int aux_mode = 0, op_code = 0, output_type = 0, address = 0, length = 0;
    unsigned char *data = NULL;
    unsigned int ret = 0;
    char*            temp1=NULL;
       
    temp = zx_calloc(count * sizeof(char));
    ret = copy_from_user(temp, buf, count);
    temp[count-1] = '\0';
    
    temp1 = zx_calloc(count * sizeof(char));
    ret = copy_from_user(temp1, buf, count);
    temp1[count-1] = '\0';

    i = 0;
    while(temp != NULL)
    {
        token = strsep(&temp, filter);
	if(token != NULL)
	{
            i++;
	}
    }

    NumOfStrings = i;

    str = (char**)zx_calloc(NumOfStrings *sizeof(char*));

    for(i = 0; i < NumOfStrings; i++)
    {
        str[i] = (char*)zx_calloc(10 * sizeof(char));
    }


    i = 0;
    while(temp1 != NULL)
    {
        token = strsep(&temp1, filter);
	if(token != NULL)
	{
            zx_strcpy(str[i], token);
	    i++;
	}
    }

    if(zx_strncmp(str[0], "native", 6) == 0)
    {
        aux_mode = 0;
        i = 1;
        while(i < NumOfStrings -1)
        {
            if(zx_strncmp(str[i], "-o", 2) == 0)
            {
                output_type = simple_strtoul(str[i+1], NULL, 0);
                i = i + 2;
            }
            else if((zx_strncmp(str[i], "-r", 2) == 0) || (zx_strncmp(str[i], "-w", 2) == 0))
            {
                op_code = (zx_strncmp(str[i], "-r", 2) == 0) ? 0 : 1;
                address = simple_strtoul(str[i+1], NULL, 0);
                i = i + 2;
            }
            else if (zx_strncmp(str[i], "-l", 2) == 0)
            {
                length = simple_strtoul(str[i+1], NULL, 0);
                i = i + 2;
                data = zx_calloc(length);
            }
            else if(zx_strncmp(str[i], "-d", 2) == 0)
            {
                for(j = 0; j < length; j++)
                {
                    data[j] =  simple_strtoul(str[i+1+j], NULL, 0);
                }
                i = i + length + 1;
            }
            else
            {
                zx_info("error input\n");
                zx_aux_help();
                goto exit;
            }	     
        }
    }
    else if(zx_strncmp(str[0], "i2c", 3) == 0)
    {
        aux_mode = 1;
        i = 1;
        while(i < NumOfStrings - 1)
        {
            if(zx_strncmp(str[i], "-o", 2) == 0)
            {
                output_type = simple_strtoul(str[i+1], NULL, 0);
                i = i + 2;
            }
            else if((zx_strncmp(str[i], "-r", 2) == 0))
            {
                op_code = 0;
                address = simple_strtoul(str[i+1], NULL, 0);
                i = i + 2;
            }
            else if (zx_strncmp(str[i], "-l", 2) == 0)
            {
                length = simple_strtoul(str[i+1], NULL, 0);
                i = i + 2;
                data = zx_calloc(length);
            }
            else
            {
                zx_info("error input\n");
                zx_aux_help();
                goto exit;
            }
        }
    }
    else
    {
        zx_aux_help();
        goto exit;
    }

//check input para
    if((output_type != OUTPUT_TYPE_DP5) && (output_type != OUTPUT_TYPE_DP6))
    {
        output_type = OUTPUT_TYPE_DP6;
        zx_info("device id input error, default is 0x10000(DP2)\n");
    }
    if(length == 0)
    {
        length = (aux_mode == 0) ? 1 : 512;
        data = zx_calloc(length);
        zx_info("input length is 0, default is 1\n");
    }

    aux.aux_mode = aux_mode;
    aux.op_code = op_code;
    aux.output_type = output_type;
    aux.address = address;
    aux.length = length;

    aux.pbuffer = zx_calloc(length);
    zx_memcpy(aux.pbuffer, data, length);

    if(aux.pbuffer != NULL)
    {
        krnl_access_aux(zx->adapter, &aux);
        if(op_code == 0)
        {
            for(i = 0; i < length; i++)
            {
                zx_info("0x%2x\n", aux.pbuffer[i]);
            }
        }
    }

exit:
    if(data != NULL)
    {
        zx_free(data);
    }
    
    for(i = 0; i < NumOfStrings; i++)
    {
        if(str[i] != NULL)
            zx_free(str[i]);
    }
    if(str != NULL)
    {
        zx_free(str);
    }
    if(temp != NULL)
    {
        zx_free(temp);
    }
    if(temp1 != NULL)
    {
        zx_free(temp1);
    }

    return count;	
}
static const struct file_operations zx_proc_fops =
{
    .owner = THIS_MODULE,
    .read  = zx_proc_read,
    .write = zx_proc_write,
};



static const struct file_operations cbios_proc_fops =
{
    .owner = THIS_MODULE,
    .read  = cbios_proc_read,
    .write = cbios_proc_write,
};

static const struct file_operations zx_aux_proc_fops =
{
    .owner = THIS_MODULE,
    .write = zx_aux_proc_write,
};

#endif

static  ssize_t  disp_dbg_level_get(struct device *dev, struct device_attribute *attr, char *buf)
{
    int  dbg_level = 0;
    void**  adapter = dev_get_drvdata(dev);
    dbg_level = krnl_get_dbg_level(*adapter);

    return sprintf(buf, "%x\n", dbg_level);
}

static  ssize_t  disp_dbg_level_set(struct device *dev, struct device_attribute *attr, const char *buf, size_t size)
{
    void**  adapter = dev_get_drvdata(dev);
    unsigned long dbg_level = 0;
    ssize_t  ret = 0;

    ret = kstrtoul(buf, 16, &dbg_level);       
    if(!ret)
    {
        krnl_set_dbg_level(*adapter, dbg_level);
        ret = size;
    }

    return ret;
}

static struct device_attribute dbg_level_attr = __ATTR(dbg_level, 0644, disp_dbg_level_get, disp_dbg_level_set);

void zx_card_pre_init(zx_card_t *zx, void *pdev)
{
#ifdef CONFIG_FB
    zx->pdev  = pdev;

    zx->zxfb_enable = zx_fb;

    zx_fb_detect_vesa(zx);
#endif

    return;
}
int zx_card_init(zx_card_t *zx, void *pdev)
{
    unsigned short command = 0;
    int ret = 0, primary, run_on_hypervisor = 0;

    zx->index = zx - zx_cards;

    spin_lock_init(&zx->wait_list_lock);
    INIT_LIST_HEAD(&zx->wait_list);

    zx_init_bus_id(zx);

    zx_get_command_status16(pdev, &command);

    primary = (command & 0x01) ? 1 : 0;

    zx_info("zx_card_init: command=0x%x, primary=%x, vesafb_enable=%d, vga16fb_enable=%d\n",
            command, primary, zx->vesafb_enable, zx->vga16fb_enable);

    /* we can enable zxfb primay and no fb driver enable by default */
    if((!primary || zx->vesafb_enable || zx->vga16fb_enable) && !zx_run_on_qt)
    {
        zx->zxfb_enable = FALSE;
    }


    /* if fb not enable, we need reserve vmem for VGA, hardcode 256K for easy */
#ifdef ZX_PCIE_BUS
    if(primary && !(zx->zxfb_enable || zx->vesafb_enable || zx->vga16fb_enable))
    {
        zx->reserved_vmem = 256 * 1024;
        zx_info("reserve 256K vmem for VGA.\n");
    }
#endif

    zx->freezable_patch = zx_freezable_patch;

    zx_info("zx->freezable_patch : 0x%x.\n", zx->freezable_patch);

    // page entry is 28bit, consider 4K(12 bit) align, set dma mask to 40(28+12).
    dma_set_mask(zx->pci_device, DMA_BIT_MASK(40));

    zx_info("query dma mask: %llx\n", dma_get_mask(zx->pci_device));

#ifdef CONFIG_X86
#ifdef CONFIG_HYPERVISOR_GUEST
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,15,0)
    if(x86_hyper)
    {
        zx_info("detect hypervisor: %s\n", x86_hyper->name);

        run_on_hypervisor = 1;
    }
#else
    if(!hypervisor_is_type(X86_HYPER_NATIVE))
    {
        zx_info("detect hypervisor: %d\n", x86_hyper_type);

        run_on_hypervisor = 1;
    }
#endif
#endif
#endif

    // keep head range memory unused for host use if driver run in virtual machine,
    // because host always use this head memory range as frame buffer, guest should
    // not touch this memory.
    if(run_on_hypervisor)
    {
        // patch reserved video memory size.
        zx->reserved_vmem += vmem_reserve_for_host;
    }

    zx->adapter = krnl_init_adapter(pdev, zx->reserved_vmem, zx->index, &zx_export, zx->pci_device, run_on_hypervisor);

    if(zx->adapter == NULL)
    {
        ret = -1;

        zx_error("init adapter failed\n");
    }

    if(zx->vesafb_enable)
    {
        krnl_notify_vesa_mode(zx->adapter, &zx->vesafb_mode);
    }

#ifdef ZX_PCIE_BUS
    zx->msi_enable = 1;
#else
    zx->msi_enable = 0;
#endif

#ifndef ZX_HW_NULL
    zx_interrupt_init(zx);
#endif

    zx->lock = zx_create_mutex();

    if(zx->zxfb_enable)
    {
#ifdef CONFIG_FB
        zx_fb_probe(zx, zx_fb_mode, zx_flip);
#endif
    }

    zx->device = device_create(zx_class, NULL, MKDEV(ZX_MAJOR, zx->index), &(zx->adapter), "%s%d", ZX_DEV_NAME, zx->index);

    if(zx->device == NULL)
    {
        ret = -1;
        zx_error("device_create failed.\n");
    }

    if(device_create_file(zx->device, &dbg_level_attr))
    {
        zx_info("Create debug level control file failed!");
    }


#ifdef NO_PROC_CREATE_FUNC
    //create_proce_read_entry is not used any more in Linux 3.10 and above. use proc_create_data instead.
    create_proc_read_entry(ZX_PROC_NAME, 0, NULL, zx_read_proc, zx);
#else
    proc_create_data(ZX_PROC_NAME, 0, NULL, &zx_proc_fops, zx);
    proc_create_data(CBIOS_PROC_NAME, 0, NULL, &cbios_proc_fops, zx);
    proc_create_data(ZX_AUX_PROC_NAME, 0, NULL, &zx_aux_proc_fops, zx);
#endif

    zx->debugfs_dev = zx_debugfs_create_device(zx);

    zx_keyboard_init(zx);
    zx_rc_dev_init(zx);
    zx_sync_init(zx);
    zx_backlight_init(pdev, zx);

    return ret;
}

int zx_card_deinit(zx_card_t *zx)
{
    if(zx->zxfb_enable)
    {
#ifdef CONFIG_FB
        zx_fb_remove(zx);
#endif
    }

    device_destroy(zx_class, MKDEV(ZX_MAJOR, zx->index));

    zx->device = NULL;

    krnl_wait_chip_idle(zx->adapter);

#ifndef ZX_HW_NULL
    zx_interrupt_deinit(zx);
#endif
    krnl_deinit_adapter(zx->adapter);

    zx->len = 0;

    if(zx->lock)
    {
        zx_destroy_mutex(zx->lock);
        zx->lock = NULL;
    }

    if (zx->sync_dev)
    {
        zx_sync_deinit(zx);
        zx->sync_dev = NULL;
    }
    if(zx->rc_dev)
    {
        zx_rc_dev_deinit(zx);
        zx->rc_dev = NULL;
    }

    if(zx->debugfs_dev)
    {
        zx_debugfs_destroy_device((zx_debugfs_device_t*)(zx->debugfs_dev));
        zx->debugfs_dev = NULL;
    }

    if(zx->backlight_dev)
    {
        zx_backlight_deinit(zx);
        zx->backlight_dev = NULL;
    }

    return 0;
}

int zx_create_allocation(int card, zx_create_allocation_t *create_data)
{
    zx_card_t   *zx = &zx_cards[card];
    unsigned int ret  = 0;

    if (zx->adapter)
    {
        ret = krnl_create_allocation(zx->adapter, create_data);
    }

    return ret;
}

void zx_destroy_allocation(int card, zx_destroy_allocation_t *destroy_data)
{
    zx_card_t *zx = &zx_cards[card];

    if (zx->adapter)
    {
        krnl_destroy_allocation(zx->adapter, destroy_data);
    }
}

int zx_lock_allocation(int card, zx_lock_allocation_t *lock_data)
{
    zx_card_t  *zx = &zx_cards[card];
    unsigned int ret = 0;

    if (zx->adapter)
    {
        ret = krnl_lock_allocation(zx->adapter, lock_data, ZX_MEM_KERNEL);
    }

    return ret;
}

void zx_unlock_allocation(int card, zx_unlock_allocation_t *unlock_data)
{
    zx_card_t *zx = &zx_cards[card];

    if (zx->adapter)
    {
        krnl_unlock_allocation(zx->adapter, unlock_data, ZX_MEM_KERNEL);
    }
}

void zx_flush_allocation(int card, zx_flush_allocation_t *flush_data)
{
    zx_card_t *zx = &zx_cards[card];

    if (zx->adapter)
    {
        krnl_flush_allocation(zx->adapter, flush_data);
    }
}

void zx_open_allocation(int card, zx_open_allocation_t *open_data)
{
    zx_card_t *zx = &zx_cards[card];

    if (zx->adapter)
    {
        krnl_open_allocation(zx->adapter, open_data);
    }
}

int zx_create_device(int card, unsigned int *device)
{
    zx_card_t  *zx = &zx_cards[card];
    unsigned int ret = 0;

    if (zx->adapter)
    {
        ret = krnl_create_device(zx->adapter, NULL, device);
    }

    return ret;
}

void zx_destroy_device(int card, unsigned int device)
{
    zx_card_t *zx = &zx_cards[card];

    if (zx->adapter)
    {
        krnl_destroy_device(zx->adapter, device);
    }
}

void zx_begin_end_isp_dvfs(int card, unsigned int on)
{
    zx_card_t *zx = &zx_cards[card];

    if (zx->adapter)
    {
        krnl_begin_end_isp_dvfs(zx->adapter, on);
    }
}

EXPORT_SYMBOL(zx_create_allocation);
EXPORT_SYMBOL(zx_destroy_allocation);
EXPORT_SYMBOL(zx_lock_allocation);
EXPORT_SYMBOL(zx_unlock_allocation);

EXPORT_SYMBOL(zx_create_device);
EXPORT_SYMBOL(zx_destroy_device);

EXPORT_SYMBOL(zx_begin_end_isp_dvfs);
