/* 
* Copyright 1998-2014 VIA Technologies, Inc. All Rights Reserved.
* Copyright 2001-2014 S3 Graphics, Inc. All Rights Reserved.
* Copyright 2013-2014 Shanghai Zhaoxin Semiconductor Co., Ltd. All Rights Reserved.
*
* This file is part of zx.ko
* 
* zx.ko is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* zx.ko is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "zx_def.h"
#include "zx_ioctl.h"
#include "os_interface.h"
#include "kernel_interface.h"
#include "zx_screen.h"
#include "zx_sync.h"
#include "zx_buffer_object.h"
#include "zx_debugfs.h"
/*****************************************************************************
**
** ioctl interface
**
*****************************************************************************/
int zx_ioctl_addctx(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    void __user           *argp = (void __user*)arg;
    zx_ctx_t              ctx;

    if(zx_copy_from_user(&ctx, argp, sizeof(ctx)))
    {
        return -1;
    }

    ctx.handle = 100;

    if(zx_copy_to_user(argp, &ctx, sizeof(ctx)))
    {
        return -1;
    }

    return 0;
}

int zx_ioctl_rmctx(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    return 0;
}

int zx_ioctl_lock_ctx(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t            *zx  = priv->card;
    //we clear freezable before lock thread, this value record the  thread real value to restore when unlock.
    int                   freezable = zx_freezable();

    if (freezable)
    {
        if ((zx->freezable_patch == 0) || (zx->freezable_patch == 1))
        {
            zx_clear_freezable();
        }
    }

    if (zx_mutex_lock_killable(zx->lock))
    {
        return -1;
    }

    priv->hold_lock = TRUE;
    priv->freezable = freezable;

    return 0;
}

int zx_ioctl_unlock_ctx(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t            *zx  = priv->card;

    priv->hold_lock = FALSE;

    zx_mutex_unlock(zx->lock);

    if(priv->freezable)
    {
        if (zx->freezable_patch == 0)
        {
            zx_set_freezable();
        }
        else if (zx->freezable_patch == 1)
        {
            zx_old_set_freezable();
        }
    }

    return 0;
}

int zx_ioctl_get_busid(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t            *zx  = priv->card;

    void __user           *argp = (void __user*)arg;
    char __user           *unique;
    zx_unique_t          u;


    if(zx_copy_from_user(&u, argp, sizeof(u)))
    {
        return -1;
    }

    if(u.unique_len >= zx->len)
    {
        unique = ptr64_to_ptr(u.unique);
        if(zx_copy_to_user(unique, zx->busId, zx->len))
        {
            return -1;
        }
    }

    u.unique_len = zx->len;

    if(zx_copy_to_user(argp, &u, sizeof(u)))
    {
        return -1;
    }

    return 0;
}

int zx_ioctl_wait_func(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t    *zx  = priv->card;
    void __user   *argp = (void __user*)arg;

    zx_wait_func_t wait;

    if(zx_copy_from_user(&wait, argp, sizeof(zx_wait_func_t)))
    {
        return -1;
    }

    switch (wait.opcode)
    {
    case ZX_WAIT_ADD:
        zx_add_wait(zx, priv, wait.events);
        break;
    case ZX_WAIT_REMOVE:
        zx_remove_wait(zx, priv);
        break;
    case ZX_WAIT_GET_EVENT:
        zx_get_event(zx, priv, &wait);
        break;
    default:
        zx_error("unknown wait opcode: %d.\n", wait.opcode);
        break;
    }

    if(zx_copy_to_user(argp, &wait, sizeof(zx_wait_func_t)))
    {
        return -1;
    }

    return 0;
}


int zx_ioctl_get_display_info(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t            *zx  = priv->card;
    void __user           *argp = (void __user*)arg;

    zx_display_info_t    display_info;

    if(zx_copy_from_user(&display_info, argp, sizeof(zx_display_info_t)))
    {
        return -1;
    }

    if(zx_get_display_info(zx, &display_info))
    {
        return -1;
    }

    if(zx_copy_to_user(argp, &display_info, sizeof(display_info)))
    {
        return -1;
    }

    return 0;
}


int zx_ioctl_get_iga_info(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t                    *zx  = priv->card;
    void __user                   *argp = (void __user*)arg;
    zx_iga_info_t                iga_info;

    if(zx_copy_from_user(&iga_info, argp, sizeof(zx_iga_info_t)))
    {
        return -1;
    }

    krnl_get_iga_info(zx->adapter, &iga_info);
    
    if(zx_copy_to_user(argp, &iga_info, sizeof(iga_info)))
    {
        return -1;
    }

    return 0;
}

static int zx_ioctl_clip_notify(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t                    *zx  = priv->card;
    void __user                   *argp = (void __user*)arg;
    zx_clipnotify_t              clip;

    if(zx_copy_from_user(&clip, argp, sizeof(clip)))
    {
        return -1;
    }

    zx_clip_notify(zx, &clip);

    return 0;
}

static int zx_ioctl_video_drawable_damage(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t                    *zx  = priv->card;
    void __user                   *argp = (void __user*)arg;
    zx_video_damage_t            damage;

    if(zx_copy_from_user(&damage, argp, sizeof(zx_video_damage_t)))
    {
        return -1;
    }

    zx_video_drawable_damage(zx, &damage);

    if(zx_copy_to_user(argp, &damage, sizeof(zx_video_damage_t)))
    {
        return -1;
    }

    return 0;
}

static int zx_ioctl_add_drawable(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t                    *zx  = priv->card;
    void __user                   *argp = (void __user*)arg;
    zx_drawable_t                drawable = {0, };

    if(zx_copy_from_user(&drawable, argp, sizeof(zx_drawable_t)))
    {
        return -1;
    }

    zx_add_drawable(zx, &drawable);

    if(zx_copy_to_user(argp, &drawable, sizeof(zx_drawable_t)))
    {
        return -1;
    }

    return 0;
}

int zx_ioctl_remove_drawable(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t                    *zx  = priv->card;
    void __user                   *argp = (void __user*)arg;
    unsigned int                  drawable_id;

    if(zx_copy_from_user(&drawable_id, argp, sizeof(drawable_id)))
    {
        return -1;
    }

    zx_remove_drawable(zx, drawable_id);

    return 0;
}

int zx_ioctl_get_drawsize(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t                    *zx  = priv->card;
    void __user                   *argp = (void __user*)arg;

    zx_get_drawsize_t            draw_rec;

    if(zx_copy_from_user(&draw_rec, argp, sizeof(draw_rec)))
    {
        return -1;
    }

    zx_get_drawable_info(zx, &draw_rec);

    if(zx_copy_to_user(argp, &draw_rec, sizeof(draw_rec)))
    {
        return -1;
    }

    return 0;
}

 int  zx_ioctl_get_device_comb(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t            *zx  = priv->card;
    void __user           *argp = (void __user*)arg;

    zx_dev_comb_t     device_comb;

    if(zx_copy_from_user(&device_comb, argp, sizeof(zx_dev_comb_t)))
    {
        return -1;
    }

    krnl_get_device_comb(zx->adapter, &device_comb);

    if(zx_copy_to_user(argp, &device_comb, sizeof(zx_dev_comb_t)))
    {
        return -1;
    }

    return 0;    
}

int  zx_ioctl_get_hpd_info(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t            *zx  = priv->card;
    void __user           *argp = (void __user*)arg;

    zx_hpd_info_t         hpd_info;

    if(zx_copy_from_user(&hpd_info, argp, sizeof(zx_hpd_info_t)))
    {
        return -1;
    }

    krnl_get_hpd_info(zx->adapter, &hpd_info);

    if(zx_copy_to_user(argp, &hpd_info, sizeof(zx_hpd_info_t)))
    {
        return -1;
    }

    return 0;    
}

int zx_ioctl_query_display_caps(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t            *zx  = priv->card;
    void __user           *argp = (void __user*)arg;

    zx_display_caps_t     display_caps;

    if(zx_copy_from_user(&display_caps, argp, sizeof(display_caps)))
    {
        return -1;
    }

    krnl_query_display_caps(zx->adapter, &display_caps);

    if(zx_copy_to_user(argp, &display_caps, sizeof(display_caps)))
    {
        return -1;
    }

    return 0;
}

static int zx_ioctl_query_output_caps(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t            *zx  = priv->card;
    void __user           *argp = (void __user*)arg;

    zx_output_caps_t  output_caps;

    if(zx_copy_from_user(&output_caps, argp, sizeof(zx_output_caps_t)))
    {
        return -1;
    }    

    krnl_query_output_caps(zx->adapter, &output_caps, 1);

    if(zx_copy_to_user(argp, &output_caps, sizeof(output_caps)))
    {
        return -1;
    }

    return 0;
}

int zx_ioctl_set_mode(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t            *zx  = priv->card;
    void __user           *argp = (void __user*)arg;

    zx_mode_info_t       mode;

    if(zx_copy_from_user(&mode, argp, sizeof(zx_mode_info_t)))
    {
        return -1;
    }

    if(mode.crtc == 0)
    {
        zx_memcpy(&zx->tv_mode, &mode, sizeof(zx_mode_info_t));
    }

    krnl_set_mode(zx->adapter, &mode);

    return 0;
}

int zx_ioctl_register_mhl_callback(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t            *zx  = priv->card;

    krnl_register_mhl_callback(zx->adapter);

    return 0;
}

int zx_ioctl_get_hdmi_audio_format(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t                    *zx  = priv->card;
    void __user                   *argp = (void __user*)arg;

    zx_hdmi_audio_formats         audio_formats = {0};

    if(krnl_get_hdmi_audio_format(zx->adapter, &audio_formats))
    {
        return -1;
    }

    if(zx_copy_to_user(argp, &audio_formats, sizeof(zx_hdmi_audio_formats)))
    {
        return -1;
    }

    return 0;
}

int zx_ioctl_cursor(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t            *zx  = priv->card;
    void __user           *argp = (void __user*)arg;

    zx_set_cursor_t       cursor_set;

    if(zx_copy_from_user(&cursor_set, argp, sizeof(zx_set_cursor_t)))
    {
        return -1;
    }

    krnl_set_cursor(zx->adapter, &cursor_set);

    return 0;
}

static int zx_ioctl_gamma(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t            *zx  = priv->card;
    void __user           *argp = (void __user*)arg;

    int                   ret   = -1;

    zx_gamma_info_t  gamma = {0};

    unsigned int* lut_buffer = zx_malloc(256 * sizeof(unsigned int));

    if(!zx_copy_from_user(&gamma, argp, sizeof(gamma)) 
       && !zx_copy_from_user(lut_buffer, gamma.pLutBuf, 256 * sizeof(unsigned int)))
    {
        gamma.pLutBuf = lut_buffer;
    
        krnl_set_gamma(zx->adapter, &gamma);

        ret = 0;
    }

    zx_free(lut_buffer);

    return ret;
}

int zx_ioctl_dpms(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t            *zx  = priv->card;
    void __user           *argp = (void __user*)arg;

    zx_dpms_set_t        dpms_set;
    
    if(zx_copy_from_user(&dpms_set, argp, sizeof(dpms_set)))
    {
        return -1;
    }

    krnl_set_dpms(zx->adapter, &dpms_set);

    return 0;
}

int zx_ioctl_get_mode_timing(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t                    *zx  = priv->card;
    void __user                   *argp = (void __user*)arg;

    zx_mode_timing_info_t        timing_info;

    if(zx_copy_from_user(&timing_info, argp, sizeof(timing_info)))
    {
        return -1;
    }

    if(krnl_get_mode_timing(zx->adapter, &timing_info))
    {
        return -1;
    }

    if(zx_copy_to_user(argp, &timing_info, sizeof(timing_info)))
    {
        return -1;
    }

    return 0;
}


int zx_ioctl_2d_init(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    void __user       *argp = (void __user*)arg;
    int               ret   = -1;

    zx_2d_init_t     init;
    struct drm_file* drm_file = priv->drm_file;
    struct file* filp = drm_file->filp;

    zx_copy_from_user(&init, argp, sizeof(init));

    switch (init.func)
    {
    case ZX_INIT:
        zx_set_server_info(priv->card, init.server_index, init.xinerama);

        init.device      = 
        priv->gpu_device = krnl_create_2ddevice(priv->card->adapter, filp, init.client_type);

        priv->need_cleanup_drawable = 1;
        priv->server_index = init.server_index;

        ret = 0;
        break;

    case ZX_CLEANUP:
        ret = krnl_destroy_2ddevice(priv->card->adapter, filp);
        priv->gpu_device = 0;
        break;

    default:
        break;

    }
    
    zx_copy_to_user(argp, &init, sizeof(init));

    return ret;
}

static int zx_ioctl_init_screen(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    void __user       *argp = (void __user*)arg;

    zx_2d_screen_t   screen;

    if(zx_copy_from_user(&screen, argp, sizeof(zx_2d_screen_t)))
    {
        return -1;
    }

    zx_init_2d_screen(priv->card, priv, &screen);

    if(zx_copy_to_user(argp, &screen, sizeof(zx_2d_screen_t)))
    {
        return -1;
    }

    return 0;
}

static int zx_ioctl_deinit_screen(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    void __user       *argp = (void __user*)arg;

    zx_2d_screen_t   screen;

    if(zx_copy_from_user(&screen, argp, sizeof(zx_2d_screen_t)))
    {
        return -1;
    }

    zx_deinit_2d_screen(priv->card, &screen);

    return 0;
}

int zx_ioctl_wait_chip_idle(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t  *zx = priv->card;

    krnl_wait_chip_idle(zx->adapter);

    return 0;
}

int _zx_get_max_covered_iga(zx_file_t *priv, unsigned int hwDrawableId);

#define MAX2(a,b)       ((a) > (b) ? (a) : (b))
#define MAX4(a,b,c,d)   MAX2(MAX2(a,b), MAX2(c,d))

#define MIN2(a,b)       ((a) > (b) ? (b) : (a))
#define MIN4(a,b,c,d)   MIN2(MIN2(a,b),MIN2(c,d))
int _zx_get_max_covered_iga(zx_file_t *priv, unsigned int hwDrawableId)
{
    zx_card_t      *zx  = priv->card;

    zx_drawable_t          *drawable = NULL;
    zx_iga_virtual_info_t  *vinfo = NULL;
    zx_clipnotify_t        *clip;
    unsigned int   iga_index;
    int   coveredx = 0, coveredy = 0, covered_size = 0, max_covered = 0;
    int   max_covered_iga = -1;

    drawable = zx_get_drawable(hwDrawableId);

    for(iga_index = 0; iga_index < MAX_CORE_CRTCS; iga_index++)
    {
        vinfo = zx->iga_virtual_info + iga_index;
        if (vinfo->active)
        {
            if (max_covered_iga == -1)
                max_covered_iga = iga_index;

            clip = &(drawable->clipNotify[0]);

            coveredx = vinfo->width + clip->width - 
                MAX4(vinfo->virtualX, vinfo->virtualX + vinfo->width, clip->screen_x, clip->screen_x + clip->width) +
                MIN4(vinfo->virtualX, vinfo->virtualX + vinfo->width, clip->screen_x, clip->screen_x + clip->width);

            coveredy = vinfo->height + clip->height - 
                MAX4(vinfo->virtualY, vinfo->virtualY + vinfo->height, clip->y, clip->screen_y + clip->height) +
                MIN4(vinfo->virtualY, vinfo->virtualY + vinfo->height, clip->y, clip->screen_y + clip->height);

            if (coveredx > 0 && coveredy > 0)
            {
                covered_size = coveredx * coveredy;
                if (covered_size > max_covered)
                {
                    max_covered = covered_size;
                    max_covered_iga = iga_index;
                }
            }
        }
    }

    return max_covered_iga;
}

int zx_ioctl_wait_event(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t  *zx = priv->card;
    void __user *argp = (void __user*)arg;
    zx_wait_func_t  wait_func_para = {0};
    int   max_covered_iga = -1;
    
    if(zx_copy_from_user(&wait_func_para, argp, sizeof(zx_wait_func_t)))
    {
        return -1;
    }

    if(wait_func_para.events == ZX_EVENT_WAIT_VSYNC_WITH_DRAWABLEID)
    {
        max_covered_iga = _zx_get_max_covered_iga(priv, wait_func_para.hwDrawableId);
        if(max_covered_iga != -1)
        {
            switch(max_covered_iga)
            {
                case 0:
                    wait_func_para.events = ZX_EVENT_VSYNC_0_INTERRUPT;
                    break;

                case 1:
                    wait_func_para.events = ZX_EVENT_VSYNC_1_INTERRUPT;
                    break;

                case 2:
                    wait_func_para.events = ZX_EVENT_VSYNC_2_INTERRUPT;
                    break;

                default:
                    break;
            }
        }
    }

    krnl_wait_event(zx->adapter, &wait_func_para);
    
    return 0;
}

static int zx_ioctl_escape_call(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t                       *zx  = priv->card;
    void __user                      *argp = (void __user*)arg;

    krnl_hint_t                      hint  = {0};
    zx_escape_call_t                escape;
    int                              ret = 0;

    if(zx_copy_from_user(&escape, argp, sizeof(zx_escape_call_t)))
    {
        return -1;
    }

    hint.filp    = priv->drm_file->filp;
    hint.hDevice = priv->gpu_device;

    ret = krnl_escape_call(zx->adapter, &hint, &escape);

    if(hint.create_device)
    {
        priv->gpu_device = hint.hDevice;
        if(zx->debugfs_dev)
        {
            zx_debugfs_add_node(zx->debugfs_dev, DEBUGFS_NODE_DEVICE, zx_get_current_pid(), priv->gpu_device, DEBUGFS_ESCAPE_CREATE);
        }
    }
    else if(hint.destroy_device)
    {
        if(zx->debugfs_dev)
        {
            zx_debugfs_remove_node(zx->debugfs_dev, priv->gpu_device);
        }

        priv->gpu_device = 0;
    }

    return ret;
}

static int zx_ioctl_2d_draw(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t                       *zx  = priv->card;
    void __user                      *argp = (void __user*)arg;

    zx_2d_draw_t                    draw_2d;

    if(zx_copy_from_user(&draw_2d, argp, sizeof(zx_2d_draw_t)))
    {
        return -1;
    }
    
    if( draw_2d.len > 0 )
    {
        draw_2d.kernel_mode_command = 0;
        krnl_kickoff_2d_draw(zx->adapter, &draw_2d);
    }

    return 0;
}

int zx_ioctl_acquire_aperture(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t                       *zx  = priv->card;
    void __user                      *argp = (void __user*)arg;
    zx_acquire_aperture_t           acquire_aperture = {0};
    
    if(zx_copy_from_user(&acquire_aperture, argp, sizeof(zx_acquire_aperture_t)))
    {
        return -1;
    }

    krnl_acquire_aperture(zx->adapter, &acquire_aperture);

    return 0;
 }

int zx_ioctl_release_aperture(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t                       *zx  = priv->card;
    void __user                      *argp = (void __user*)arg;
    zx_release_aperture_t           release_aperture = {0};
    
    if(zx_copy_from_user(&release_aperture, argp, sizeof(zx_release_aperture_t)))
    {
        return -1;
    }

    krnl_release_aperture(zx->adapter, &release_aperture);

    return 0;
}

int zx_ioctl_leave_vt(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t   *zx  = priv->card;

    zx->enter_vt = FALSE;

    krnl_leave_zx_graphics_mode(zx->adapter, priv->drm_file->filp);

#ifdef NEW_ZXFB
    if(zx->info && zx->zxfb_pending_events)
    {
        unsigned int events = zx->zxfb_pending_events;

        zx->zxfb_pending_events = 0;

        zxfb_handle_event(zx, events);
    }
#endif

    return 0;
}

int zx_ioctl_enter_vt(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t   *zx  = priv->card;

    zx->enter_vt = TRUE;

    krnl_enter_zx_graphics_mode(zx->adapter, priv->drm_file->filp);

    return 0;
}

int zx_ioctl_check_device_combination(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t                       *zx  = priv->card;
    void __user                      *argp = (void __user*)arg;
    zx_check_device_combination_t   device_combination;

    if(zx_copy_from_user(&device_combination, argp, sizeof(device_combination)))
    {
        return -1;
    }

    if(krnl_check_device_combination(zx->adapter, &device_combination))
    {
        return -1;
    }

    if(zx_copy_to_user(argp, &device_combination, sizeof(device_combination)))
    {
        return -1;
    }

    return 0;
}

int zx_ioctl_set_ps(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t        *zx  = priv->card;
    void __user       *argp = (void __user*)arg;
    framebuffer_t     *fb   = NULL;
    zx_ps_info_t     ps;
    
    if(zx_copy_from_user(&ps, argp, sizeof(zx_ps_info_t)))
    {
        return -1;
    }

    fb = zx_get_frame_buffer(zx, ps.server_index, ps.screen_index);

    if(fb != NULL)
    {
        fb->rot_allocation[ps.crtc] = ps.shadow_surface;
    }

    krnl_set_ps(zx->adapter, &ps);

    return 0;
}

int zx_ioctl_flip(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t        *zx  = priv->card;
    void __user       *argp = (void __user*)arg;
    zx_flip_info_t   flip;
    zx_create_fence_fd_t create_fence = {0};

    if(zx_copy_from_user(&flip, argp, sizeof(zx_flip_info_t)))
    {
        return -1;
    }

    flip.retire_fence_fd = -1;
    flip.retire_fence_object = 0;

    if(flip.need_retire_fence && flip.hDevice)
    {
        create_fence.fence_value   = 1;
        create_fence.device        = flip.hDevice;

        if(zx_sync_fence_create_fd(zx->sync_dev, &create_fence) == 0)
        {
            flip.retire_fence_fd        = create_fence.fd;
            flip.retire_fence_object    = create_fence.sync_obj;
        }
    }

    krnl_flip(zx->adapter, &flip);

    if (zx_copy_to_user(argp, &flip, sizeof(zx_flip_info_t)))
    {
        return -1;
    }
    return 0;
}

int zx_ioctl_enum_sub_device(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    int status = -1;

    if (krnl_enum_sub_device(priv->card->adapter) == S_OK)
    {
        status = 0;
    }

    return status;
}

int zx_ioctl_create_allocation(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t        *zx  = priv->card;
    void __user       *argp = (void __user*)arg;

    zx_create_allocation_t create_data;
    int                     ret = 0;

    if(zx_copy_from_user(&create_data, argp, sizeof(zx_create_allocation_t)))
    {
        return -1;
    }

    if(create_data.need_fd)
    {
        //LeonHe:
        //zx_buffer_oject_create() is not used in e2uma
        zx_assert(0);

        //ret = zx_buffer_object_create(zx->adapter, &create_data);
        if(ret < 0)
        {
            return -1;
        }
    }
    else
    {
        ret = krnl_create_allocation(zx->adapter, &create_data);
    }

    if(zx_copy_to_user(argp, &create_data, sizeof(zx_create_allocation_t)))
    {
        return -1;
    }

    return ret;
}

int zx_ioctl_destroy_allocation(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t        *zx  = priv->card;
    void __user       *argp = (void __user*)arg;

    zx_destroy_allocation_t destroy_data;

    if(zx_copy_from_user(&destroy_data, argp, sizeof(zx_destroy_allocation_t)))
    {
        return -1;
    }

    krnl_destroy_allocation(zx->adapter, &destroy_data);

    return 0;
}

static int zx_ioctl_lock_allocation(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t        *zx  = priv->card;
    void __user       *argp = (void __user*)arg;

    zx_lock_allocation_t lock_data;
    int                   ret;

    if(zx_copy_from_user(&lock_data, argp, sizeof(zx_lock_allocation_t)))
    {
        return -1;
    }

    ret = krnl_lock_allocation(zx->adapter, &lock_data, ZX_MEM_USER);

    if(zx_copy_to_user(argp, &lock_data, sizeof(zx_lock_allocation_t)))
    {
        return -1;
    }

    return ret;
}

int zx_ioctl_unlock_allocation(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t        *zx  = priv->card;
    void __user       *argp = (void __user*)arg;

    zx_unlock_allocation_t unlock_data;

    if(zx_copy_from_user(&unlock_data, argp, sizeof(zx_unlock_allocation_t)))
    {
        return -1;
    }

    krnl_unlock_allocation(zx->adapter, &unlock_data, ZX_MEM_USER);

    if(zx_copy_to_user(argp, &unlock_data, sizeof(zx_unlock_allocation_t)))
    {
        return -1;
    }

    return 0;
}

int zx_ioctl_flush_allocation(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t        *zx  = priv->card;
    void __user       *argp = (void __user*)arg;

    zx_flush_allocation_t flush_data;

    if(zx_copy_from_user(&flush_data, argp, sizeof(zx_flush_allocation_t)))
    {
        return -1;
    }

    krnl_flush_allocation(zx->adapter, &flush_data);

    return 0;
}

static int zx_ioctl_ipc_operation(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t              *zx  = priv->card;
    void __user             *argp = (void __user*)arg;
    int                     ret=0;

    zx_ipc_operation_t     ipc_operation;

    if(zx_copy_from_user(&ipc_operation, argp, sizeof(zx_ipc_operation_t)))
    {
        return -1;
    }

    ret = zx_ipc_operation(zx, priv, &ipc_operation);

    return ret;
}

int zx_ioctl_set_interactive(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t        *zx = priv->card;
    void __user       *argp = (void __user*)arg;
    int    active = 0;

    if(zx_copy_from_user(&active, argp, sizeof(int)))
    {
        return -1;
    }

    krnl_set_interactive(zx->adapter, active);
    return  0;
}

int zx_ioctl_open_allocation(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t        *zx  = priv->card;
    void __user       *argp = (void __user*)arg;

    zx_open_allocation_t open_data;
    int                   ret;

    if(zx_copy_from_user(&open_data, argp, sizeof(zx_open_allocation_t)))
    {
        return -1;
    }

    ret = krnl_open_allocation(zx->adapter, &open_data);

    if(zx_copy_to_user(argp, &open_data, sizeof(zx_open_allocation_t)))
    {
        return -1;
    }

    return ret;
}

int zx_ioctl_wait_allocation_idle(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t                 *zx  = priv->card;
    void __user                *argp = (void __user*)arg;

    zx_wait_allocation_idle_t  wait_allocation;

    if(zx_copy_from_user(&wait_allocation, argp, sizeof(zx_wait_allocation_idle_t)))
    {
        return -1;
    }

    krnl_wait_allocation_idle(zx->adapter, &wait_allocation);

    return 0;
}

int zx_ioctl_video_get_hw_caps(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t  *zx  = priv->card;
    void __user *argp = (void __user*)arg;

    unsigned char caps = krnl_get_video_hardware_capability(zx->adapter);

    if(zx_copy_to_user(argp, &caps, sizeof(unsigned char)))
    {
        return -1;
    }

    return 0;
}

static int zx_ioctl_get_address_config(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t  *zx  = priv->card;
    void __user *argp = (void __user*)arg;

    zx_get_address_config_t address_config;

    krnl_get_address_config(zx->adapter, &address_config);

    if(zx_copy_to_user(argp, &address_config, sizeof(zx_get_address_config_t)))
    {
        return -1;
    }

    return 0;
}

int zx_ioctl_query_info(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t  *zx  = priv->card;
    void __user *argp = (void __user*)arg;
    zx_query_info_t info;
    int ret = 0;
    
    if(zx_copy_from_user(&info, argp, sizeof(zx_query_info_t)))
    {
        return -1;
    }

    ret = krnl_query_info(zx->adapter, &info);

    if(zx_copy_to_user(argp, &info, sizeof(zx_query_info_t)))
    {
        return -1;
    }

    return ret;
}

int zx_ioctl_create_stream(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t      *zx  = priv->card;
    void __user     *argp = (void __user*)arg;
    zx_create_stream_t  create_stream;
    int ret = 0;

    if(zx_copy_from_user(&create_stream, argp, sizeof(zx_create_stream_t)))
    {
        return -1;
    }

    ret =  krnl_create_stream(zx->adapter, &create_stream);

    if(zx_copy_to_user(argp, &create_stream, sizeof(zx_create_stream_t)))
    {
        return -1;
    }

    return ret;
}

int zx_ioctl_destroy_stream(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t      *zx  = priv->card;
    void __user     *argp = (void __user*)arg;
    zx_destroy_stream_t  destroy_stream;
    int ret = 0;

    if(zx_copy_from_user(&destroy_stream, argp, sizeof(zx_destroy_stream_t)))
    {
        return -1;
    }

    krnl_destroy_stream(zx->adapter, &destroy_stream);

    if(zx_copy_to_user(argp, &destroy_stream, sizeof(zx_destroy_stream_t)))
    {
        return -1;
    }

    return ret;
}

int zx_ioctl_display_frame(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t            *zx          = priv->card;
    void __user           *argp         = (void __user*)arg;
    zx_display_frame_t   display_frame;
    zx_create_fence_fd_t create_fence  = {0};

    int ret = 0;

    if(zx_copy_from_user(&display_frame, argp, sizeof(zx_display_frame_t)))
    {
        return -1;
    }

    //active_stream_mask should be subset of acquired_stream_mask
    if(display_frame.active_stream_mask != (display_frame.active_stream_mask & display_frame.acquired_stream_mask))
    {
        zx_error("zxDisplayFrame invalid mask:%x-%x\n", display_frame.active_stream_mask, display_frame.acquired_stream_mask);
        return -1;
    }

    display_frame.retire_fence_fd     = -1;
    display_frame.retire_fence_object = 0;

    if(display_frame.need_retire_fence && display_frame.hDevice)
    {
        create_fence.fence_value = ZX_DISPLAY_RETIRE_SIGNAL_VALUE;
        create_fence.device      = display_frame.hDevice;
        create_fence.context     = display_frame.hContext;

        if(zx_sync_fence_create_fd(zx->sync_dev, &create_fence) == 0)
        {
            display_frame.retire_fence_fd     = create_fence.fd;
            display_frame.retire_fence_object = create_fence.sync_obj;
        }
    }

    display_frame.release_fence_fd     = -1;
    display_frame.release_fence_object = 0;

    if(display_frame.need_release_fence && display_frame.hDevice)
    {
        create_fence.fence_value = ZX_DISPLAY_RELEASE_SIGNAL_VALUE;
        create_fence.device      = display_frame.hDevice;

        if(zx_sync_fence_create_fd(zx->sync_dev, &create_fence) == 0)
        {
            display_frame.release_fence_fd     = create_fence.fd;
            display_frame.release_fence_object = create_fence.sync_obj;
        }
    }

    ret =  krnl_display_frame(zx->adapter, &display_frame);

    if(zx_copy_to_user(argp, &display_frame, sizeof(zx_display_frame_t)))
    {
        return -1;
    }

 //   krnl_free_vesafb_tempbuffer(zx->adapter);

    return ret;
}

int zx_ioctl_cec_ctl(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t      *zx  = priv->card;
    void __user     *argp = (void __user*)arg;
    zx_cec_ctl_t   cec_ctl;
    int ret = 0;

    if(zx_copy_from_user(&cec_ctl, argp, sizeof(zx_cec_ctl_t)))
    {
        return -1;
    }

    zx_info("zx ioctl cec ctl\n");

    ret = krnl_cec_ctl(zx->adapter, &cec_ctl);

    if(zx_copy_to_user(argp, &cec_ctl, sizeof(zx_cec_ctl_t)))
    {
        return -1;
    }

    return ret;
}

int zx_ioctl_present(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t      *zx  = priv->card;
    void __user     *argp = (void __user*)arg;

    zx_drawable_t  *drawable = NULL;
    zx_present_t   present;

    unsigned int  last = 0, scrn_index, status = 0;

    if(zx_copy_from_user(&present, argp, sizeof(zx_present_t)))
    {
        return -1;
    }

    if(present.ignore_drawable)
    {
        zx_clipnotify_t clip = {0, };
        zx_clip_rect_t stack_rects[8];
        zx_clip_rect_t *clip_rect = stack_rects;

        if (present.num_cliprects > 0)
        {
            if (present.num_cliprects > sizeof(stack_rects)/sizeof(stack_rects[0]))
            {
                clip_rect = zx_malloc(sizeof(*clip_rect) * present.num_cliprects);
            }
            zx_copy_from_user(clip_rect, present.cliprects, present.num_cliprects * sizeof(*clip_rect));

            clip.hAllocation = present.dst_allocation;
            clip.clipRects = clip_rect;
            clip.numClipRects = present.num_cliprects;
        }
        else
        {
            framebuffer_t *fb = zx_get_frame_buffer(zx, present.server_index, present.screen_index);

            if (present.dst_allocation)
            {
                /* this case means blit src_allocation to dst_allocation */
                clip.hAllocation = present.dst_allocation;
            }
            else
            {
                /* this case means flip fb->hAllocation to diu */
                clip.hAllocation = fb->hAllocation;
            }
            clip_rect->x1 = 0;
            clip_rect->y1 = 0;
            clip_rect->x2 = fb->width;
            clip_rect->y2 = fb->height;
            clip.clipRects = clip_rect;
            clip.numClipRects = 1;
        }

        status = krnl_present(zx->adapter, &present, &clip);

        if (clip_rect != stack_rects)
        {
            zx_free(clip_rect);
        }

        return status;
    }

    drawable  = zx_get_drawable(present.drawable_id);

    for(scrn_index = 0; scrn_index < MAX_SCREENS; scrn_index++)
    {
        if(drawable->clipNotify[scrn_index].hAllocation 
           && last != drawable->clipNotify[scrn_index].hAllocation)
        {
            framebuffer_t *fb = zx_get_frame_buffer(priv->card, present.server_index, present.screen_index);

            if (fb && fb->adapter == zx->adapter)
            {
                status = krnl_present(zx->adapter, &present, &drawable->clipNotify[scrn_index]);
                last   = drawable->clipNotify[scrn_index].hAllocation;
            }

            if(status != 0)
            {
                break;
            }
        }
    }

    return status;
}

int zx_ioctl_get_drawable_surface(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t      *zx  = priv->card;
    void __user     *argp = (void __user*)arg;

    zx_get_drawable_surface_t   get_surface;

    unsigned int    hAllocation = 0;

    if(zx_copy_from_user(&get_surface, argp, sizeof(zx_get_drawable_surface_t)))
    {
        return -1;
    }

    /* drawable_id 0 means get primary surface */
    if(get_surface.drawable_id == 0)
    {
        framebuffer_t *fb = zx_get_frame_buffer(zx, get_surface.server_index, get_surface.screen_index);

        hAllocation = fb->hAllocation;
    }
    else
    {
        zx_drawable_t *drawable = zx_get_drawable(get_surface.drawable_id);

        hAllocation = drawable->clipNotify[get_surface.screen_index].hAllocation;
    }

    get_surface.surface = hAllocation;

    if(zx_copy_to_user(argp, &get_surface, sizeof(zx_get_drawable_surface_t)))
    {
        return -1;
    }

    return 0;
}

int zx_ioctl_get_vsync_info(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t      *zx  = priv->card;
    void __user     *argp = (void __user*)arg;

    zx_get_vsync_info_t     get_vsync;
    int   max_covered_iga = -1;

    if(zx_copy_from_user(&get_vsync, argp, sizeof(zx_get_vsync_info_t)))
    {
        return -1;
    }

    max_covered_iga = _zx_get_max_covered_iga(priv, get_vsync.drawable_id);

    if(max_covered_iga == -1)
    {
        return -1;
    }

    get_vsync.vsync_cnt = zx->v_sync_count[max_covered_iga];
    get_vsync.vsync_interval = krnl_get_iga_vsync_interval(zx->adapter, max_covered_iga);

    if(zx_copy_to_user(argp, &get_vsync, sizeof(zx_get_vsync_info_t)))
    {
        return -1;
    }

    return 0;
}

int zx_ioctl_create_device(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t            *zx = priv->card;
    void __user           *argp = (void __user*)arg;

    zx_create_device_t    create_device = {0};
    int                    ret  = 0;

    zx_assert(priv->gpu_device == 0);

    ret = krnl_create_device(zx->adapter, priv->drm_file->filp, &create_device.device);

    if(zx_copy_to_user(argp, &create_device, sizeof(zx_create_device_t)))
    {
        return -1;
    }

    priv->gpu_device = create_device.device;

    if(zx->debugfs_dev)
    {
        zx_debugfs_add_node(zx->debugfs_dev, DEBUGFS_NODE_DEVICE, zx_get_current_pid(), priv->gpu_device, DEBUGFS_ZX_CREATE);
    }

    return ret;
}

int zx_ioctl_destroy_device(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t        *zx = priv->card;
    void __user       *argp = (void __user*)arg;

    unsigned int       device = 0;


    if(zx_copy_from_user(&device, argp, sizeof(unsigned int)))
    {
        return -1;
    }

    zx_info("destroy device %x.\n", device);

    if(zx->debugfs_dev)
    {
        zx_debugfs_remove_node(zx->debugfs_dev, priv->gpu_device);
    }

    krnl_destroy_device(zx->adapter, device);

    priv->gpu_device = 0;

    return 0;
}

int zx_ioctl_begin_perf_event(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t   *zx = priv->card;
    void __user  *argp = (void __user*)arg;
    zx_begin_perf_event_t begin_perf_event = {0};
    int          ret = 0;

    if (zx_copy_from_user(&begin_perf_event, argp, sizeof(zx_begin_perf_event_t)))
    {
        return -1;
    }

    ret = krnl_begin_perf_event(zx->adapter, &begin_perf_event);

    return ret;
}

int zx_ioctl_begin_miu_dump_perf_event(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t *zx = priv->card;
    void __user *argp = (void __user*)arg;
    zx_begin_miu_dump_perf_event_t begin_miu_dump = {0};
    int ret = 0;

    if(zx_copy_from_user(&begin_miu_dump, argp, sizeof(zx_begin_miu_dump_perf_event_t)))
    {
        return -1;
    }

    ret = krnl_begin_miu_dump_perf_event(zx->adapter, &begin_miu_dump);

    return ret;
}

int zx_ioctl_end_perf_event(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t   *zx = priv->card;
    void __user  *argp = (void __user*)arg;
    zx_end_perf_event_t end_perf_event = {0};
    int          ret = 0;

    if (zx_copy_from_user(&end_perf_event, argp, sizeof(zx_end_perf_event_t)))
    {
        return -1;
    }

    ret = krnl_end_perf_event(zx->adapter, &end_perf_event);

    return ret;
}

int zx_ioctl_end_miu_dump_perf_event(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t *zx = priv->card;
    void __user *argp = (void __user*)arg;
    zx_end_miu_dump_perf_event_t end_miu_dump = {0};
    int ret = 0;

    if(zx_copy_from_user(&end_miu_dump, argp, sizeof(zx_end_miu_dump_perf_event_t)))
    {
        return -1;
    }

    ret = krnl_end_miu_dump_perf_event(zx->adapter, &end_miu_dump);

    return ret;
}


static int zx_ioctl_set_miu_reg_list_perf_event(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t *zx = priv->card;
    void __user *argp = (void __user*)arg;
    zx_miu_reg_list_perf_event_t miu_reg_list = {0};
    int ret = 0;

    if(zx_copy_from_user(&miu_reg_list, argp, sizeof(zx_miu_reg_list_perf_event_t)))
    {
        return -1;
    }

    ret = krnl_set_miu_reg_list_perf_event(zx->adapter, &miu_reg_list);

    return ret;
}

static int zx_ioctl_get_miu_dump_perf_event(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t *zx = priv->card;
    void __user *argp = (void __user*)arg;
    zx_get_miu_dump_perf_event_t get_miu_dump = {0};
    int ret = 0;

    if(zx_copy_from_user(&get_miu_dump, argp, sizeof(zx_get_miu_dump_perf_event_t)))
    {
        return -1;
    }

    ret = krnl_get_miu_dump_perf_event(zx->adapter, &get_miu_dump);

    if (ret == 0)
    {
        zx_copy_to_user(argp, &get_miu_dump, sizeof(zx_get_miu_dump_perf_event_t));
    }

    return ret;
}

static int zx_ioctl_direct_get_miu_dump_perf_event(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t *zx = priv->card;
    void __user *argp = (void __user*)arg;
    zx_direct_get_miu_dump_perf_event_t direct_get_dump = {0};
    int ret = 0;

    if(zx_copy_from_user(&direct_get_dump, argp, sizeof(zx_direct_get_miu_dump_perf_event_t)))
    {
        return -1;
    }

    ret = krnl_direct_get_miu_dump_perf_event(zx->adapter, &direct_get_dump);

    if(ret == 0)
    {
        zx_copy_to_user(argp, &direct_get_dump, sizeof(zx_direct_get_miu_dump_perf_event_t));
    }

    return ret;
}

static int zx_ioctl_get_perf_event(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t   *zx = priv->card;
    void __user  *argp = (void __user*)arg;
    zx_get_perf_event_t get_perf_event = {0};
    int          ret = 0;

    if (zx_copy_from_user(&get_perf_event, argp, sizeof(zx_get_perf_event_t)))
    {
        return -1;
    }

    ret = krnl_get_perf_event(zx->adapter, &get_perf_event);

    if (ret == 0)
    {
        zx_copy_to_user(argp, &get_perf_event, sizeof(zx_get_perf_event_t));
    }

    return ret;
}

int zx_ioctl_send_perf_event(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t   *zx = priv->card;
    void __user  *argp = (void __user*)arg;
    zx_perf_event_t perf_event;
    int          ret = 0;

    if (zx_copy_from_user(&perf_event, argp, sizeof(zx_perf_event_header_t)))
    {
        return -1;
    }

    if (perf_event.header.size > sizeof(zx_perf_event_header_t))
    {
        if (zx_copy_from_user(&perf_event, argp, perf_event.header.size))
        {
            return -1;
        }
    }

    ret = krnl_send_perf_event(zx->adapter, &perf_event);

    return ret;
}


int zx_ioctl_get_perf_status(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t   *zx = priv->card;
    void __user  *argp = (void __user*)arg;
    zx_perf_status_t perf_status = {0};
    int ret = 0;

    ret = krnl_get_perf_status(zx->adapter, &perf_status);

    if (ret == 0)
    {
        zx_copy_to_user(argp, &perf_status, sizeof(zx_perf_status_t));
    }

    return ret;
}

static int zx_ioctl_create_context(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t          *zx  = priv->card;
    void __user         *argp = (void __user*)arg;

    zx_create_context_t create_context = {0};
    int                  ret  = 0;

    if(zx_copy_from_user(&create_context, argp, sizeof(zx_create_context_t)))
    {
        return -1;
    }

    ret = krnl_create_context(zx->adapter, &create_context, ZX_MEM_USER);

    if (ret != S_OK)
    {
        return -1;
    }

    if(zx_copy_to_user(argp, &create_context, sizeof(zx_create_context_t)))
    {
        return -1;
    }

    return ret;
}

int zx_ioctl_destroy_context(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t          *zx  = priv->card;
    void __user         *argp = (void __user*)arg;

    zx_destroy_context_t destroy_context = {0};
    int                   ret  = 0;

    if(zx_copy_from_user(&destroy_context, argp, sizeof(zx_destroy_context_t)))
    {
        return -1;
    }

    ret = krnl_destroy_context(zx->adapter, &destroy_context);

    if(zx_copy_to_user(argp, &destroy_context, sizeof(zx_destroy_context_t)))
    {
        return -1;
    }

    return ret;
}

int zx_ioctl_create_di_context(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t          *zx  = priv->card;
    void __user         *argp = (void __user*)arg;

    zx_create_di_context_t create_context;
    int                  ret  = 0;

    if(zx_copy_from_user(&create_context, argp, sizeof(zx_create_di_context_t)))
    {
        return -1;
    }

    ret = krnl_create_di_context(zx->adapter, &create_context);

    if(zx_copy_to_user(argp, &create_context, sizeof(zx_create_di_context_t)))
    {
        return -1;
    }

    return ret;
}

int zx_ioctl_destroy_di_context(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t          *zx  = priv->card;
    void __user         *argp = (void __user*)arg;

    zx_destroy_di_context_t destroy_context;
    int                      ret  = 0;

    if(zx_copy_from_user(&destroy_context, argp, sizeof(zx_destroy_di_context_t)))
    {
        return -1;
    }

    ret = krnl_destroy_di_context(zx->adapter, &destroy_context);

    if(zx_copy_to_user(argp, &destroy_context, sizeof(zx_destroy_di_context_t)))
    {
        return -1;
    }

    return ret;
}

static int zx_ioctl_render(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t   *zx    = priv->card;
    void __user  *argp   = (void __user*)arg;

    zx_render_t  render = {0};
    int           ret    = 0;

    if(zx_copy_from_user(&render, argp, sizeof(zx_render_t)))
    {
        return -1;
    }

    ret = krnl_render(zx->adapter, &render);

    if(zx_copy_to_user(argp, &render, sizeof(zx_render_t)))
    {
        return -1;
    }

    return ret;
}

int zx_ioctl_query_chip_id(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t        *zx  = priv->card;
    void __user       *argp = (void __user*)arg;
    unsigned int      ret = 0;

    zx_chip_id_t     chip_id;
    zx_query_info_t  query;

    if(zx_copy_from_user(&chip_id, argp, sizeof(zx_chip_id_t)))
    {
        return -1;
    }

    query.type = ZX_QUERY_VENDOR_ID;

    krnl_query_info(zx->adapter, &query);

    chip_id.Vendor_ID = query.value;

    query.type = ZX_QUERY_DEVICE_ID;

    krnl_query_info(zx->adapter, &query);

    chip_id.Device_ID = query.value >> 8;

    query.type = ZX_QUERY_REVISION_ID;

    krnl_query_info(zx->adapter, &query);

    chip_id.Revision_ID = query.value;

    if(zx_copy_to_user(argp, &chip_id, sizeof(zx_chip_id_t)))
    {
        return -1;
    }

    return ret;
}

int zx_ioctl_hdcp_op(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t        *zx  = priv->card;
    void __user       *argp = (void __user*)arg;
    unsigned int      ret;

    zx_hdcp_op_t     hdcp_op;

    if(zx_copy_from_user(&hdcp_op, argp, sizeof(zx_hdcp_op_t)))
    {
        return -1;
    }

    ret = krnl_hdcp_op(zx->adapter, priv, &hdcp_op);

    if(zx_copy_to_user(argp, &hdcp_op, sizeof(zx_hdcp_op_t)))
    {
        return -1;
    }

    return ret;
}

int zx_ioctl_create_fence_sync_object(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t   *zx    = priv->card;
    void __user  *argp   = (void __user*)arg;
    int          result  = 0;

    zx_create_fence_sync_object_t  create = {0};

    if(zx_copy_from_user(&create, argp, sizeof(zx_create_fence_sync_object_t)))
    {
        return -1;
    }

    result = krnl_create_fence_sync_object(zx->adapter, &create, FALSE);

    if(zx_copy_to_user(argp, &create, sizeof(zx_create_fence_sync_object_t)))
    {
        return -1;
    }

    return result;
}

int zx_ioctl_destroy_fence_sync_object(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t   *zx    = priv->card;
    void __user  *argp   = (void __user*)arg;
    int          result  = 0;

    zx_destroy_fence_sync_object_t  destroy = {0};

    if(zx_copy_from_user(&destroy, argp, sizeof(zx_destroy_fence_sync_object_t)))
    {
        return -1;
    }

    result = krnl_destroy_fence_sync_object(zx->adapter, &destroy);

    return result;
}

int zx_ioctl_wait_fence_sync_object(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t   *zx    = priv->card;
    void __user  *argp   = (void __user*)arg;
    int          result  = 0;

    zx_wait_fence_sync_object_t  wait = {0};

    if(zx_copy_from_user(&wait, argp, sizeof(zx_wait_fence_sync_object_t)))
    {
        return -1;
    }

    result = krnl_wait_fence_sync_object(zx->adapter, &wait);

    if(zx_copy_to_user(argp, &wait, sizeof(zx_wait_fence_sync_object_t)))
    {
        return -1;
    }

    return result;
}

int zx_ioctl_fence_value(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t   *zx    = priv->card;
    void __user  *argp   = (void __user*)arg;
    int          result  = 0;

    zx_fence_value_t  value = {0};

    if(zx_copy_from_user(&value, argp, sizeof(zx_fence_value_t)))
    {
        return -1;
    }

    result = krnl_fence_value(zx->adapter, &value);

    if(zx_copy_to_user(argp, &value, sizeof(zx_fence_value_t)))
    {
        return -1;
    }

    return result;
}

int zx_ioctl_create_fence_fd(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t   *zx    = priv->card;
    void __user  *argp   = (void __user*)arg;
    int          result  = 0;

    zx_create_fence_fd_t value = {0};

    if(zx_copy_from_user(&value, argp, sizeof(zx_create_fence_fd_t)))
    {
        return -1;
    }

    result = zx_sync_fence_create_fd(zx->sync_dev, &value);

    if(zx_copy_to_user(argp, &value, sizeof(zx_create_fence_fd_t)))
    {
        return -1;
    }

    return result;

}

int zx_ioctl_flip_limit(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t            *zx  = priv->card;
    void __user           *argp = (void __user*)arg;

    zx_flip_limit_t    flipvalue;

    if(zx_copy_from_user(&flipvalue, argp, sizeof(zx_flip_limit_t)))
    {
        return -1;
    }

    krnl_flip_limit(zx->adapter, &flipvalue);

   return 0;
}

int zx_ioctl_wait_context_idle(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t        *zx  = priv->card;
    void __user       *argp = (void __user*)arg;

    unsigned int context = 0;

    if(zx_copy_from_user(&context, argp, sizeof(unsigned int)))
    {
        return -1;
    }

    krnl_wait_context_idle(zx->adapter, context);

    return 0;
}

static int zx_ioctl_send_dsi_cmd(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t      *zx  = priv->card;
    void __user     *argp = (void __user*)arg;
    zx_dsi_cmd_t dsi_cmd;
    int ret = 0;

    if(zx_copy_from_user(&dsi_cmd, argp, sizeof(zx_dsi_cmd_t)))
    {
        return -1;
    }

    ret = krnl_send_dsi_cmd(zx->adapter, &dsi_cmd);

    if(zx_copy_to_user(argp, &dsi_cmd, sizeof(zx_dsi_cmd_t)))
    {
        return -1;
    }

    return ret;
}

int zx_ioctl_mhl_cbus_send_command(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t      *zx  = priv->card;
    void __user     *argp = (void __user*)arg;
    zx_cbus_cmd_t cbus_cmd;
    int ret = 0;

    if(zx_copy_from_user(&cbus_cmd, argp, sizeof(zx_cbus_cmd_t)))
    {
        return -1;
    }

    ret = krnl_mhl_cbus_send_command(zx->adapter, &cbus_cmd);

    if(zx_copy_to_user(argp, &cbus_cmd, sizeof(zx_cbus_cmd_t)))
    {
        return -1;
    }

    return ret;
}

int zx_ioctl_set_power_perf_mode(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t            *zx  = priv->card;
    void __user           *argp = (void __user*)arg;

    zx_power_perf_mode_info_t       mode;

    if(zx_copy_from_user(&mode, argp, sizeof(zx_power_perf_mode_info_t)))
    {
        return -1;
    }

    krnl_set_power_perf_mode(zx->adapter, &mode);

    return 0;
}

int zx_ioctl_gfx_stm(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t      *zx  = priv->card;
    void __user     *argp = (void __user*)arg;
    int             is_resume = 0;
    int ret = 0;

    if(zx_copy_from_user(&is_resume, argp, sizeof(int)))
    {
        return -1;
    }

    if(is_resume)
    {
        zx_interrupt_reinit(zx);
        ret = krnl_restore_state(zx->adapter);
    }
    else
    {
        krnl_save_state(zx->adapter, 1);
        zx_interrupt_deinit(zx);
    }

    if(zx_copy_to_user(argp, &is_resume, sizeof(is_resume)))
    {
        return -1;
    }

    return ret;
}

int zx_ioctl_fd_to_handle(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    int             ret;
    void __user     *argp = (void __user*)arg;
    zx_handle_fd_t hfd;

    if(zx_copy_from_user(&hfd, argp, sizeof(zx_handle_fd_t)))
    {
        return -1;
    }

    hfd.handle = 0;

    ret = zx_bo_fd_to_handle(priv, &hfd);
    if (ret)
    {
        return ret;
    }

    if(zx_copy_to_user(argp, &hfd, sizeof(zx_handle_fd_t)))
    {
        return -1;
    }

    return 0;
}

int zx_ioctl_handle_to_fd(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    int             ret = 0;
    void __user     *argp = (void __user*)arg;
    zx_handle_fd_t hfd;

    if(zx_copy_from_user(&hfd, argp, sizeof(zx_handle_fd_t)))
    {
        return -1;
    }

    hfd.fd = -1;

    ret = zx_bo_handle_to_fd(priv, &hfd);

    if (ret)
    {
        return ret;
    }

    if(zx_copy_to_user(argp, &hfd, sizeof(zx_handle_fd_t)))
    {
        return -1;
    }
    return 0;
}

int zx_ioctl_get_ust_msc(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t        *zx = priv->card;
    void __user       *argp = (void __user*)arg;
    zx_ust_msc_t     um = {0, };

    if(zx_copy_from_user(&um, argp, sizeof(zx_ust_msc_t)))
    {
        return -1;
    }

    if (um.crtc < 0 || um.crtc >= sizeof(zx->v_sync_count) / sizeof(zx->v_sync_count[0]))
    {
        return -1;
    }

    um.msc = zx->v_sync_count[um.crtc];
    um.ust = zx_do_div(zx->v_sync_timestamp[um.crtc] , 1000);

    if(zx_copy_to_user(argp, &um, sizeof(zx_ust_msc_t)))
    {
        return -1;
    }

    return 0;
}

int zx_ioctl_iga_virtual_info(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t        *zx = priv->card;
    void __user       *argp = (void __user*)arg;
    zx_iga_virtual_info_t info = {0, };

    if (zx_copy_from_user(&info, argp, sizeof(info)))
    {
        return -1;
    }

    if (info.iga_index < 0 || info.iga_index >= sizeof(zx->iga_virtual_info) / sizeof(zx->iga_virtual_info[0]))
    {
        return -1;
    }

    zx_memcpy(zx->iga_virtual_info + info.iga_index, &info, sizeof(info));

    return 0;
}

int zx_ioctl_get_allocation_state(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    int               ret = -1;
    zx_card_t        *zx = priv->card;
    void __user       *argp = (void __user*)arg;
    zx_get_allocation_state_t state = {0, };

    if (zx_copy_from_user(&state, argp, sizeof(state)))
    {
        return -1;
    }

    ret = krnl_get_allocation_state(zx->adapter, &state);

    if (ret == 0)
    {
        ret = zx_copy_to_user(argp, &state, sizeof(state));
    }

    return ret;
}

int zx_ioctl_access_aux(zx_file_t *priv, unsigned int cmd, unsigned long arg)
{
    zx_card_t        *zx = priv->card;
    void __user *argp = (void __user*)arg;
    zx_aux_t aux_para;
    int ret = 0;
    
    if(zx_copy_from_user(&aux_para, argp, sizeof(zx_aux_t)))
    {
        return -1;
    }

    ret = krnl_access_aux(zx->adapter, &aux_para);

    if(zx_copy_to_user(argp, &aux_para, sizeof(zx_aux_t)))
    {
        return -1;
    }

    return ret;
}

zx_ioctl_t zx_ioctls[] =
{
    [ZX_IOCTL_NR(ZX_IOCTL_GET_BUSID)]                = zx_ioctl_get_busid,
    [ZX_IOCTL_NR(ZX_IOCTL_GET_DISPLAYINFO)]          = zx_ioctl_get_display_info,
    [ZX_IOCTL_NR(ZX_IOCTL_GET_IGAINFO)]              = zx_ioctl_get_iga_info,
    [ZX_IOCTL_NR(ZX_IOCTL_ADD_CTX)]                  = zx_ioctl_addctx,
    [ZX_IOCTL_NR(ZX_IOCTL_RM_CTX)]                   = zx_ioctl_rmctx,
    [ZX_IOCTL_NR(ZX_IOCTL_LOCK)]                     = zx_ioctl_lock_ctx,
    [ZX_IOCTL_NR(ZX_IOCTL_UNLOCK)]                   = zx_ioctl_unlock_ctx,
    [ZX_IOCTL_NR(ZX_IOCTL_CLIP_NOTIFY)]              = zx_ioctl_clip_notify,
    [ZX_IOCTL_NR(ZX_IOCTL_ADD_DRAW)]                 = zx_ioctl_add_drawable,
    [ZX_IOCTL_NR(ZX_IOCTL_RM_DRAW)]                  = zx_ioctl_remove_drawable,
    [ZX_IOCTL_NR(ZX_IOCTL_GET_DRAWSIZE)]             = zx_ioctl_get_drawsize,
    [ZX_IOCTL_NR(ZX_IOCTL_QUERY_DISPLAY_CAPS)]       = zx_ioctl_query_display_caps,
    [ZX_IOCTL_NR(ZX_IOCTL_QUERY_OUTPUT_CAPS)]        = zx_ioctl_query_output_caps,
    [ZX_IOCTL_NR(ZX_IOCTL_MODE_SET)]                 = zx_ioctl_set_mode,
    [ZX_IOCTL_NR(ZX_IOCTL_CURSOR)]                   = zx_ioctl_cursor,
    [ZX_IOCTL_NR(ZX_IOCTL_GAMMA)]                    = zx_ioctl_gamma,
    [ZX_IOCTL_NR(ZX_IOCTL_DPMS)]                     = zx_ioctl_dpms,
    [ZX_IOCTL_NR(ZX_IOCTL_GET_MODE_TIMING)]          = zx_ioctl_get_mode_timing,
    [ZX_IOCTL_NR(ZX_IOCTL_2D_INIT)]                  = zx_ioctl_2d_init,
    [ZX_IOCTL_NR(ZX_IOCTL_INIT_SCREEN)]              = zx_ioctl_init_screen,
    [ZX_IOCTL_NR(ZX_IOCTL_DEINIT_SCREEN)]            = zx_ioctl_deinit_screen,    
    [ZX_IOCTL_NR(ZX_IOCTL_WAIT_CHIP_IDLE)]           = zx_ioctl_wait_chip_idle,
    [ZX_IOCTL_NR(ZX_IOCTL_ESCAPE_CALL)]              = zx_ioctl_escape_call,
    [ZX_IOCTL_NR(ZX_IOCTL_2D_DRAW)]                  = zx_ioctl_2d_draw,
    [ZX_IOCTL_NR(ZX_IOCTL_ACQUIRE_APERTURE)]         = zx_ioctl_acquire_aperture,
    [ZX_IOCTL_NR(ZX_IOCTL_RELEASE_APERTURE)]         = zx_ioctl_release_aperture,
    [ZX_IOCTL_NR(ZX_IOCTL_LEAVE_VT)]                 = zx_ioctl_leave_vt,
    [ZX_IOCTL_NR(ZX_IOCTL_ENTER_VT)]                 = zx_ioctl_enter_vt,
    [ZX_IOCTL_NR(ZX_IOCTL_CHECK_DEVICE_COMBINATION)] = zx_ioctl_check_device_combination,
    [ZX_IOCTL_NR(ZX_IOCTL_SET_PS)]                   = zx_ioctl_set_ps,
    [ZX_IOCTL_NR(ZX_IOCTL_ENUM_SUB_DEVICE)]          = zx_ioctl_enum_sub_device,
    [ZX_IOCTL_NR(ZX_IOCTL_CREATE_ALLOCATION)]        = zx_ioctl_create_allocation,
    [ZX_IOCTL_NR(ZX_IOCTL_OPEN_ALLOCATION)]          = zx_ioctl_open_allocation,
    [ZX_IOCTL_NR(ZX_IOCTL_DESTROY_ALLOCATION)]       = zx_ioctl_destroy_allocation,
    [ZX_IOCTL_NR(ZX_IOCTL_LOCK_ALLOCATION)]          = zx_ioctl_lock_allocation,
    [ZX_IOCTL_NR(ZX_IOCTL_UNLOCK_ALLOCATION)]        = zx_ioctl_unlock_allocation,
    [ZX_IOCTL_NR(ZX_IOCTL_FLUSH_ALLOCATION)]         = zx_ioctl_flush_allocation,
    [ZX_IOCTL_NR(ZX_IOCTL_WAIT_ALLOCATION_IDLE)]     = zx_ioctl_wait_allocation_idle,
    [ZX_IOCTL_NR(ZX_IOCTL_VIDEO_GET_HW_CAPS)]        = zx_ioctl_video_get_hw_caps,
    [ZX_IOCTL_NR(ZX_IOCTL_GET_ADDRESS_CONFIG)]       = zx_ioctl_get_address_config,
    [ZX_IOCTL_NR(ZX_IOCTL_QUERY_INFO)]               = zx_ioctl_query_info,
    [ZX_IOCTL_NR(ZX_IOCTL_WAIT_FUNC)]                = zx_ioctl_wait_func,
    [ZX_IOCTL_NR(ZX_IOCTL_FLIP)]                     = zx_ioctl_flip,
    [ZX_IOCTL_NR(ZX_IOCTL_CEC_CTL)]                  = zx_ioctl_cec_ctl,
    [ZX_IOCTL_NR(ZX_IOCTL_PRESENT)]                  = zx_ioctl_present,
    [ZX_IOCTL_NR(ZX_IOCTL_GET_DRAWABLE_SURFACE)]     = zx_ioctl_get_drawable_surface,
    [ZX_IOCTL_NR(ZX_IOCTL_GET_VSYNC_INFO)]           = zx_ioctl_get_vsync_info,
    [ZX_IOCTL_NR(ZX_IOCTL_CREATE_DEVICE)]            = zx_ioctl_create_device,
    [ZX_IOCTL_NR(ZX_IOCTL_DESTROY_DEVICE)]           = zx_ioctl_destroy_device,
    [ZX_IOCTL_NR(ZX_IOCTL_BEGIN_PERF_EVENT)]         = zx_ioctl_begin_perf_event,
    [ZX_IOCTL_NR(ZX_IOCTL_END_PERF_EVENT)]           = zx_ioctl_end_perf_event,
    [ZX_IOCTL_NR(ZX_IOCTL_GET_PERF_EVENT)]           = zx_ioctl_get_perf_event,
    [ZX_IOCTL_NR(ZX_IOCTL_SEND_PERF_EVENT)]          = zx_ioctl_send_perf_event,
    [ZX_IOCTL_NR(ZX_IOCTL_GET_PERF_STATUS)]          = zx_ioctl_get_perf_status,
    [ZX_IOCTL_NR(ZX_IOCTL_BEGIN_MIU_DUMP_PERF_EVENT)] = zx_ioctl_begin_miu_dump_perf_event,
    [ZX_IOCTL_NR(ZX_IOCTL_END_MIU_DUMP_PERF_EVENT)]  = zx_ioctl_end_miu_dump_perf_event,
    [ZX_IOCTL_NR(ZX_IOCTL_SET_MIU_REG_LIST_PERF_EVENT)] = zx_ioctl_set_miu_reg_list_perf_event,
    [ZX_IOCTL_NR(ZX_IOCTL_GET_MIU_DUMP_PERF_EVENT)]  = zx_ioctl_get_miu_dump_perf_event,
    [ZX_IOCTL_NR(ZX_IOCTL_DIRECT_GET_MIU_DUMP_PERF_EVENT)] = zx_ioctl_direct_get_miu_dump_perf_event,
    [ZX_IOCTL_NR(ZX_IOCTL_CREATE_CONTEXT)]           = zx_ioctl_create_context,
    [ZX_IOCTL_NR(ZX_IOCTL_DESTROY_CONTEXT)]          = zx_ioctl_destroy_context,
    [ZX_IOCTL_NR(ZX_IOCTL_WAIT_CONTEXT_IDLE)]        = zx_ioctl_wait_context_idle,
    [ZX_IOCTL_NR(ZX_IOCTL_RENDER)]                   = zx_ioctl_render,
    [ZX_IOCTL_NR(ZX_IOCTL_QUERY_CHIP_ID)]            = zx_ioctl_query_chip_id,
    [ZX_IOCTL_NR(ZX_IOCTL_HDCP_OP)]                  = zx_ioctl_hdcp_op,
    [ZX_IOCTL_NR(ZX_IOCTL_CREATE_FENCE_SYNC_OBJECT)] = zx_ioctl_create_fence_sync_object,
    [ZX_IOCTL_NR(ZX_IOCTL_DESTROY_FENCE_SYNC_OBJECT)]= zx_ioctl_destroy_fence_sync_object,
    [ZX_IOCTL_NR(ZX_IOCTL_WAIT_FENCE_SYNC_OBJECT)]   = zx_ioctl_wait_fence_sync_object,
    [ZX_IOCTL_NR(ZX_IOCTL_FENCE_VALUE)]              = zx_ioctl_fence_value,
    [ZX_IOCTL_NR(ZX_IOCTL_CREATE_FENCE_FD)]          = zx_ioctl_create_fence_fd,
    [ZX_IOCTL_NR(ZX_IOCTL_SET_FLIP_COUNTER)]         = zx_ioctl_flip_limit,
    [ZX_IOCTL_NR(ZX_IOCTL_IPC_OPERATION)]            = zx_ioctl_ipc_operation,
    [ZX_IOCTL_NR(ZX_IOCTL_CREATE_STREAM)]            = zx_ioctl_create_stream,
    [ZX_IOCTL_NR(ZX_IOCTL_DESTROY_STREAM)]           = zx_ioctl_destroy_stream,
    [ZX_IOCTL_NR(ZX_IOCTL_DISPLAY_FRAME)]            = zx_ioctl_display_frame,
    [ZX_IOCTL_NR(ZX_IOCTL_SET_INTERACTIVE)]          = zx_ioctl_set_interactive,
    [ZX_IOCTL_NR(ZX_IOCTL_SEND_DSI_CMD)]             = zx_ioctl_send_dsi_cmd,
    [ZX_IOCTL_NR(ZX_IOCTL_MHL_CBUS_SEND_CMD)]        = zx_ioctl_mhl_cbus_send_command,
    [ZX_IOCTL_NR(ZX_IOCTL_CREATE_DI_CONTEXT)]        = zx_ioctl_create_di_context,
    [ZX_IOCTL_NR(ZX_IOCTL_DESTROY_DI_CONTEXT)]       = zx_ioctl_destroy_di_context,
    [ZX_IOCTL_NR(ZX_IOCTL_REGISTER_MHL_CALLBACK)]    = zx_ioctl_register_mhl_callback,
    [ZX_IOCTL_NR(ZX_IOCTL_SET_POWER_PERF_MODE)]      = zx_ioctl_set_power_perf_mode,
    [ZX_IOCTL_NR(ZX_IOCTL_GET_DEVICE_COMB)]          = zx_ioctl_get_device_comb,
    [ZX_IOCTL_NR(ZX_IOCTL_GFX_STM)]                  = zx_ioctl_gfx_stm,
    [ZX_IOCTL_NR(ZX_IOCTL_GET_HDMI_AUDIO_FORMAT)]    = zx_ioctl_get_hdmi_audio_format,
    [ZX_IOCTL_NR(ZX_IOCTL_GET_HPD_INFO)]             = zx_ioctl_get_hpd_info,
    [ZX_IOCTL_NR(ZX_IOCTL_HANDLE_TO_FD)]             = zx_ioctl_handle_to_fd,
    [ZX_IOCTL_NR(ZX_IOCTL_FD_TO_HANDLE)]             = zx_ioctl_fd_to_handle,
    [ZX_IOCTL_NR(ZX_IOCTL_GET_UST_MSC)]              = zx_ioctl_get_ust_msc,
    [ZX_IOCTL_NR(ZX_IOCTL_GET_ALLOCATION_STATE)]     = zx_ioctl_get_allocation_state,
    [ZX_IOCTL_NR(ZX_IOCTL_WAIT_EVENT)]               = zx_ioctl_wait_event,
    [ZX_IOCTL_NR(ZX_IOCTL_VIDEO_DAMAGE)]             = zx_ioctl_video_drawable_damage,
    [ZX_IOCTL_NR(ZX_IOCTL_IGA_VIRTUAL_INFO)]         = zx_ioctl_iga_virtual_info,
    [ZX_IOCTL_NR(ZX_IOCTL_ACCESS_AUX)]               = zx_ioctl_access_aux,
};




